#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import operator
from typing import TYPE_CHECKING, Callable

import numpy as np
import pandas as pd

from saqc.constants import BAD
from saqc.core.register import flagging
from saqc.lib.tools import isunflagged, statPass

if TYPE_CHECKING:
    from saqc import SaQC


class NoiseMixin:
    @flagging()
    def flagByStatLowPass(
        self: "SaQC",
        field: str,
        func: Callable[[np.ndarray, pd.Series], float],
        window: str | pd.Timedelta,
        thresh: float,
        sub_window: str | pd.Timedelta | None = None,
        sub_thresh: float | None = None,
        min_periods: int | None = None,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag data chunks of length ``window``, if:

        1. they excexceed ``thresh`` with regard to ``func`` and
        2. all (maybe overlapping) sub-chunks of the data chunks with length ``sub_window``,
           exceed ``sub_thresh`` with regard to ``func``

        Parameters
        ----------
        func :
            Aggregation function applied on every chunk.

        window :
            Window (i.e. chunk) size.

        thresh :
            Threshold. A given chunk is flagged, if the return value of ``func`` excceeds ``thresh``.

        sub_window :
            Window size of sub chunks, that are additionally tested for exceeding ``sub_thresh``
            with respect to ``func``.

        sub_thresh :
            Threshold. A given sub chunk is flagged, if the return value of ``func` excceeds ``sub_thresh``.

        min_periods :
            Minimum number of values needed in a chunk to perfom the test.
            Ignored if ``window`` is an integer.
        """

        datcol = self._data[field]
        if not min_periods:
            min_periods = 0
        if not sub_thresh:
            sub_thresh = thresh
        window = pd.Timedelta(window)

        if sub_window is not None:
            sub_window = pd.Timedelta(sub_window)

        to_set = statPass(
            datcol,
            func,
            window,
            thresh,
            operator.gt,
            sub_window,
            sub_thresh,
            min_periods,
        )
        mask = isunflagged(self._flags[field], kwargs["dfilter"]) & to_set
        self._flags[mask, field] = flag
        return self



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from copy import copy as shallowcopy
from copy import deepcopy
from typing import Any, Hashable, MutableMapping

import numpy as np
import pandas as pd

from saqc.core.flags import Flags, _HistAccess, initFlagsLike
from saqc.core.frame import DictOfSeries
from saqc.core.history import History
from saqc.core.register import FUNC_MAP
from saqc.core.translation import (
    DmpScheme,
    FloatScheme,
    PositionalScheme,
    SimpleScheme,
    TranslationScheme,
)
from saqc.funcs import FunctionsMixin

# warnings
pd.set_option("mode.chained_assignment", "warn")
pd.options.mode.copy_on_write = False
np.seterr(invalid="ignore")


TRANSLATION_SCHEMES = {
    "float": FloatScheme,
    "simple": SimpleScheme,
    "dmp": DmpScheme,
    "positional": PositionalScheme,
}


class SaQC(FunctionsMixin):
    _attributes = {
        "_data",
        "_flags",
        "_scheme",
        "_attrs",
    }

    def __init__(
        self,
        data=None,
        flags=None,
        scheme: str | TranslationScheme = "float",
    ):
        self._data: DictOfSeries = self._initData(data)
        self._flags: Flags = self._initFlags(flags)
        self._scheme: TranslationScheme = self._initTranslationScheme(scheme)
        self._attrs: dict = {}
        self._validate(reason="init")

    def _construct(self, **attributes) -> "SaQC":
        """
        Construct a new `SaQC`-Object from `self` and optionally inject
        attributes with any chechking and overhead.

        Parameters
        ----------
        **attributes: any of the `SaQC` data attributes with name and value

        Note
        ----
        For internal usage only! Setting values through `injectables` has
        the potential to mess up certain invariants of the constructed object.
        """
        out = SaQC(data=DictOfSeries(), flags=Flags(), scheme=self._scheme)
        out.attrs = self._attrs
        for k, v in attributes.items():
            if k not in self._attributes:
                raise AttributeError(f"SaQC has no attribute {repr(k)}")
            setattr(out, k, v)
        return out

    def _validate(self, reason=None):
        if not self._data.columns.equals(self._flags.columns):
            msg = "Consistency broken. data and flags have not the same columns."
            if reason:
                msg += f" This was most likely caused by: {reason}"
            raise RuntimeError(msg)

    @property
    def attrs(self) -> dict[Hashable, Any]:
        """
        Dictionary of global attributes of this dataset.
        """
        return self._attrs

    @attrs.setter
    def attrs(self, value: dict[Hashable, Any]) -> None:
        self._attrs = dict(value)

    @property
    def data(self) -> MutableMapping:
        data = self._data
        data.attrs = self._attrs.copy()
        return data

    @property
    def flags(self) -> MutableMapping:
        flags = self._scheme.toExternal(self._flags, attrs=self._attrs)
        flags.attrs = self._attrs.copy()
        return flags

    @property
    def _history(self) -> _HistAccess:
        return self._flags.history

    def __getattr__(self, key):
        """
        All failing attribute accesses are redirected to __getattr__.
        We use this mechanism to make the registered functions appear
        as `SaQC`-methods without actually implementing them.
        """
        from functools import partial

        if key not in FUNC_MAP:
            raise AttributeError(f"SaQC has no attribute {repr(key)}")
        return partial(FUNC_MAP[key], self)

    def copy(self, deep=True):
        copyfunc = deepcopy if deep else shallowcopy
        new = self._construct()
        for attr in self._attributes:
            setattr(new, attr, copyfunc(getattr(self, attr)))
        return new

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memodict=None):
        return self.copy(deep=True)

    def _initTranslationScheme(
        self, scheme: str | TranslationScheme
    ) -> TranslationScheme:
        if isinstance(scheme, str) and scheme in TRANSLATION_SCHEMES:
            return TRANSLATION_SCHEMES[scheme]()
        if isinstance(scheme, TranslationScheme):
            return scheme
        raise TypeError(
            f"expected one of the following translation schemes '{TRANSLATION_SCHEMES.keys()} "
            f"or an initialized Translator object, got '{scheme}'"
        )

    def _initData(self, data) -> DictOfSeries:
        if data is None:
            return DictOfSeries()
        if isinstance(data, list):
            result = DictOfSeries()
            doubles = pd.Index([])
            for d in data:
                new = self._castData(d)
                doubles = doubles.union(result.columns.intersection(new.columns))
                result.update(new)
            if not doubles.empty:
                warnings.warn(
                    f"Column(s) {doubles.tolist()} was present multiple "
                    f"times in input data. Some data was overwritten. "
                    f"Avoid duplicate columns names over all inputs.",
                    stacklevel=2,
                )
            return result
        try:
            return self._castData(data)
        except ValueError as e:
            raise e from None
        except TypeError as e:
            raise TypeError(
                "'data' must be of type pandas.Series, "
                "pandas.DataFrame or saqc.DictOfSeries or "
                "a list of those or a dict with string keys "
                "and pandas.Series as values."
            ) from e

    def _castData(self, data) -> DictOfSeries:
        if isinstance(data, pd.Series):
            if not isinstance(data.name, str):
                raise ValueError(f"Cannot init from unnamed pd.Series")
            data = data.to_frame()
        if isinstance(data, pd.DataFrame):
            for idx in [data.index, data.columns]:
                if isinstance(idx, pd.MultiIndex):
                    raise ValueError("'data' should not have MultiIndex")
        try:
            # This ensures that values are pd.Series
            return DictOfSeries(data)
        except Exception:
            raise TypeError(f"Cannot cast {type(data)} to DictOfSeries") from None

    def _initFlags(self, flags) -> Flags:
        if flags is None:
            return initFlagsLike(self._data)

        if isinstance(flags, list):
            result = Flags()
            for f in flags:
                f = self._castToFlags(f)
                for c in f.columns:
                    if c in result.columns:
                        warnings.warn(
                            f"Column {c} already exist. Data is overwritten. "
                            f"Avoid duplicate columns names over all inputs.",
                            stacklevel=2,
                        )
                        result.history[c] = f.history[c]
            flags = result

        elif isinstance(flags, (pd.DataFrame, DictOfSeries, Flags)):
            flags = self._castToFlags(flags)

        else:
            raise TypeError(
                "'flags' must be of type pandas.DataFrame, "
                "dios.DictOfSeries or saqc.Flags or "
                "a list of those."
            )

        # sanitize
        # - if column is missing flags but present in data, add it
        # - if column is present in both, the index must be equal
        for c in self._data.columns:
            if c not in flags.columns:
                flags.history[c] = History(self._data[c].index)
            else:
                if not flags[c].index.equals(self._data[c].index):
                    raise ValueError(
                        f"The flags index of column {c} does not equals "
                        f"the index of the same column in data."
                    )
        return flags

    def _castToFlags(self, flags):
        if isinstance(flags, pd.DataFrame):
            for idx in [flags.index, flags.columns]:
                if isinstance(idx, pd.MultiIndex):
                    raise TypeError("'flags' should not have MultiIndex")
        if not isinstance(flags, Flags):
            flags = Flags(flags)
        return flags



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import logging
from functools import partial
from pathlib import Path

import click
import numpy as np
import pandas as pd
import pyarrow as pa

from saqc.core import DictOfSeries
from saqc.core.core import TRANSLATION_SCHEMES
from saqc.parsing.reader import fromConfig

logger = logging.getLogger("SaQC")


def _setupLogging(loglvl):
    logger.setLevel(loglvl)
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s][%(name)s][%(levelname)s]: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def setupIO(nodata):
    reader = {
        ".csv": partial(pd.read_csv, index_col=0, parse_dates=True),
        ".parquet": pd.read_parquet,
    }

    writer = {
        ".csv": partial(pd.DataFrame.to_csv, header=True, index=True, na_rep=nodata),
        ".parquet": lambda df, outfile: pa.parquet.write_table(
            pa.Table.from_pandas(df), outfile
        ),
    }
    return reader, writer


def readData(reader_dict, fname):
    extension = Path(fname).suffix
    reader = reader_dict.get(extension)
    if not reader:
        raise ValueError(
            f"Unsupported file format '{extension}', use one of {tuple(reader_dict.keys())}"
        )
    return reader(fname)


def writeData(writer_dict, df, fname):
    extension = Path(fname).suffix
    writer = writer_dict.get(extension)
    if not writer:
        raise ValueError(
            f"Unsupported file format '{extension}', use one of {tuple(writer_dict.keys())}"
        )
    writer(df, fname)


@click.command()
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    required=True,
    help="path to the configuration file",
)
@click.option(
    "-d",
    "--data",
    type=click.Path(),
    multiple=True,
    required=True,
    help="path to the data file",
)
@click.option(
    "-o", "--outfile", type=click.Path(exists=False), help="path to the output file"
)
@click.option(
    "--scheme",
    default=None,
    type=click.Choice(tuple(TRANSLATION_SCHEMES.keys())),
    help="the flagging scheme to use",
)
@click.option("--nodata", default=np.nan, help="nodata value")
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING"]),
    help="set output verbosity",
)
def main(config, data, scheme, outfile, nodata, log_level):
    # data is always a list of data files

    _setupLogging(log_level)
    reader, writer = setupIO(nodata)

    data = [readData(reader, f) for f in data]

    saqc = fromConfig(
        config,
        data=data,
        scheme=TRANSLATION_SCHEMES[scheme or "simple"](),
    )

    data_result = saqc.data.to_pandas()
    flags_result = saqc.flags
    if isinstance(flags_result, DictOfSeries):
        flags_result = flags_result.to_pandas()

    if outfile:
        data_result.columns = pd.MultiIndex.from_product(
            [data_result.columns.tolist(), ["data"]]
        )

        if not isinstance(flags_result.columns, pd.MultiIndex):
            flags_result.columns = pd.MultiIndex.from_product(
                [flags_result.columns.tolist(), ["flags"]]
            )

        out = pd.concat([data_result, flags_result], axis=1).sort_index(
            axis=1, level=0, sort_remaining=False
        )

        writeData(writer, out, outfile)


if __name__ == "__main__":
    main()



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-


import numpy as np
import pandas as pd

# see test/functs/fixtures.py for global fixtures "course_..."
import pytest

import saqc
from saqc import UNFLAGGED, SaQC
from saqc.core import DictOfSeries, initFlagsLike
from saqc.lib.ts_operators import linearInterpolation, polynomialInterpolation
from tests.fixtures import char_dict, course_3, course_5  # noqa, todo: fix fixtures


def test_rollingInterpolateMissing(course_5):
    data, characteristics = course_5(periods=10, nan_slice=[5, 6])
    field = data.columns[0]
    data = DictOfSeries(data)
    flags = initFlagsLike(data)
    qc = SaQC(data, flags).interpolateByRolling(
        field,
        3,
        func=np.median,
        center=True,
        min_periods=0,
        interpol_flag=UNFLAGGED,
    )
    assert qc.data[field][characteristics["missing"]].notna().all()
    qc = SaQC(data, flags).interpolateByRolling(
        field,
        3,
        func=np.nanmean,
        center=False,
        min_periods=3,
        interpol_flag=UNFLAGGED,
    )
    assert qc.data[field][characteristics["missing"]].isna().all()


def test_interpolate(course_5):
    data, characteristics = course_5(periods=10, nan_slice=[5])
    field = data.columns[0]
    data = DictOfSeries(data)
    flags = initFlagsLike(data)
    qc = SaQC(data, flags)

    qc_lin = qc.interpolate(field, method="linear")
    qc_poly = qc.interpolate(field, method="polynomial")
    assert qc_lin.data[field][characteristics["missing"]].notna().all()
    assert qc_poly.data[field][characteristics["missing"]].notna().all()

    data, characteristics = course_5(periods=10, nan_slice=[5, 6, 7])

    qc = SaQC(data, flags)
    qc_lin_1 = qc.interpolate(field, method="linear", limit=2)
    qc_lin_2 = qc.interpolate(field, method="linear", limit=3)
    qc_lin_3 = qc.interpolate(field, method="linear", limit=4)

    assert qc_lin_1.data[field][characteristics["missing"]].isna().all()
    assert qc_lin_2.data[field][characteristics["missing"]].isna().all()
    assert qc_lin_3.data[field][characteristics["missing"]].notna().all()


def test_transform(course_5):
    data, characteristics = course_5(periods=10, nan_slice=[5, 6])
    field = data.columns[0]
    data = DictOfSeries(data)
    flags = initFlagsLike(data)
    qc = SaQC(data, flags)

    result = qc.transform(field, func=linearInterpolation)
    assert result.data[field][characteristics["missing"]].isna().all()

    result = qc.transform(field, func=lambda x: linearInterpolation(x, inter_limit=3))
    assert result.data[field][characteristics["missing"]].notna().all()

    result = qc.transform(
        field,
        func=lambda x: polynomialInterpolation(x, inter_limit=3, inter_order=3),
    )
    assert result.data[field][characteristics["missing"]].notna().all()


def test_resample(course_5):
    data, _ = course_5(freq="1min", periods=30, nan_slice=[1, 11, 12, 22, 24, 26])
    field = data.columns[0]
    data = DictOfSeries(data)
    flags = initFlagsLike(data)
    qc = SaQC(data, flags).resample(
        field,
        "10min",
        np.mean,
        maxna=2,
        maxna_group=1,
    )
    assert ~np.isnan(qc.data[field].iloc[0])
    assert np.isnan(qc.data[field].iloc[1])
    assert np.isnan(qc.data[field].iloc[2])


def test_interpolateGrid(course_5, course_3):
    data, _ = course_5()
    data_grid, _ = course_3()
    data["grid"] = data_grid["data"]
    flags = initFlagsLike(data)
    SaQC(data, flags).align("data", "1h", "time", grid_field="grid", limit=10)


@pytest.mark.slow
def test_offsetCorrecture():
    data = pd.Series(0, index=pd.date_range("2000", freq="1d", periods=100), name="dat")
    data.iloc[30:40] = -100
    data.iloc[70:80] = 100
    flags = initFlagsLike(data)
    qc = SaQC(data, flags).correctOffset("dat", 40, 20, "3d", 1)
    assert (qc.data["dat"] == 0).all()


# GL-333
def test_resampleSingleEmptySeries():
    qc = saqc.SaQC(pd.DataFrame(1, columns=["a"], index=pd.DatetimeIndex([])))
    qc.resample("a", freq="1d")


@pytest.mark.parametrize(
    "data",
    [
        pd.Series(
            [
                np.random.normal(loc=1 + k * 0.1, scale=3 * (1 - (k * 0.001)))
                for k in range(100)
            ],
            index=pd.date_range("2000", freq="1D", periods=100),
            name="data",
        )
    ],
)
def test_assignZScore(data):
    qc = saqc.SaQC(data)
    qc = qc.assignZScore("data", window="20D")
    mean_res = qc.data["data"].mean()
    std_res = qc.data["data"].std()
    assert -0.1 < mean_res < 0.1
    assert 0.9 < std_res < 1.1



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

from __future__ import annotations

import json
from functools import reduce

import numpy as np
import pandas as pd

from saqc import BAD, DOUBTFUL, GOOD, UNFLAGGED
from saqc.core import Flags, History
from saqc.core.translation.basescheme import BackwardMap, ForwardMap, MappingScheme
from saqc.lib.tools import getUnionIndex

_QUALITY_CAUSES = [
    "",
    "BATTERY_LOW",
    "BELOW_MINIMUM",
    "ABOVE_MAXIMUM",
    "BELOW_OR_ABOVE_MIN_MAX",
    "ISOLATED_SPIKE",
    "DEFECTIVE_SENSOR",
    "LEFT_CENSORED_DATA",
    "RIGHT_CENSORED_DATA",
    "OTHER",
    "AUTOFLAGGED",
]

_QUALITY_LABELS = [
    "quality_flag",
    "quality_cause",
    "quality_comment",
]


class DmpScheme(MappingScheme):

    """
    Implements the translation from and to the flagging scheme implemented in
    the UFZ - Datamanagementportal
    """

    ARGUMENTS = {"comment": "", "cause": "OTHER"}

    DFILTER_DEFAULT = GOOD + 1

    _FORWARD: ForwardMap = {
        "NIL": UNFLAGGED,
        "OK": GOOD,
        "DOUBTFUL": DOUBTFUL,
        "BAD": BAD,
    }

    _BACKWARD: BackwardMap = {
        UNFLAGGED: "NIL",
        np.nan: "NIL",
        GOOD: "OK",
        DOUBTFUL: "DOUBTFUL",
        BAD: "BAD",
    }

    def __init__(self):
        super().__init__(forward=self._FORWARD, backward=self._BACKWARD)

    def toHistory(self, field_flags: pd.DataFrame):
        """
        Translate a single field of external ``Flags`` to a ``History``
        """
        field_history = History(field_flags.index)

        for (flag, cause, comment), values in field_flags.groupby(_QUALITY_LABELS):
            try:
                comment = json.loads(comment)
            except json.decoder.JSONDecodeError:
                comment = {"test": "unknown", "comment": ""}

            histcol = pd.Series(np.nan, index=field_flags.index)
            histcol.loc[values.index] = self(flag)

            meta = {
                "func": comment["test"],
                "kwargs": {"comment": comment["comment"], "cause": cause},
            }
            field_history.append(histcol, meta=meta)
        return field_history

    def toInternal(self, df: pd.DataFrame) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        df : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """

        self.validityCheck(df)

        data = {}

        for field in df.columns.get_level_values(0).drop_duplicates():
            data[str(field)] = self.toHistory(df[field])

        return Flags(data)

    def toExternal(
        self, flags: Flags, attrs: dict | None = None, **kwargs
    ) -> pd.DataFrame:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : The external flags to translate

        attrs : dict or None, default None
            global meta information of saqc-object

        Returns
        -------
        translated flags
        """
        tflags = super().toExternal(flags, attrs=attrs)

        out = pd.DataFrame(
            index=getUnionIndex(tflags),
            columns=pd.MultiIndex.from_product([flags.columns, _QUALITY_LABELS]),
        )

        for field in tflags.columns:
            df = pd.DataFrame(
                {
                    "quality_flag": tflags[field],
                    "quality_cause": "",
                    "quality_comment": "",
                }
            )

            history = flags.history[field]
            for col in history.columns:
                valid = (history.hist[col] != UNFLAGGED) & history.hist[col].notna()

                # extract from meta
                meta = history.meta[col]
                keywords = meta.get("kwargs", {})
                comment = json.dumps(
                    {
                        "test": meta.get("func", "unknown"),
                        "comment": keywords.get("comment", self.ARGUMENTS["comment"]),
                    }
                )
                cause = keywords.get("cause", self.ARGUMENTS["cause"])
                df.loc[valid, "quality_comment"] = comment
                df.loc[valid, "quality_cause"] = cause

            out[field] = df.reindex(out.index)

        self.validityCheck(out)
        return out

    @classmethod
    def validityCheck(cls, df: pd.DataFrame) -> None:
        """
        Check wether the given causes and comments are valid.

        Parameters
        ----------
        df : external flags
        """

        cols = df.columns
        if not isinstance(cols, pd.MultiIndex):
            raise TypeError("DMP-Flags need multi-index columns")

        if not cols.get_level_values(1).isin(_QUALITY_LABELS).all(axis=None):
            raise TypeError(
                f"DMP-Flags expect the labels {list(_QUALITY_LABELS)} in the secondary level"
            )

        for field in df.columns.get_level_values(0):
            # we might have NaN injected by DictOfSeries -> DataFrame conversions
            field_df = df[field].dropna(how="all", axis="index")
            flags = field_df["quality_flag"]
            causes = field_df["quality_cause"]
            comments = field_df["quality_comment"]

            if not flags.isin(cls._FORWARD.keys()).all(axis=None):
                raise ValueError(
                    f"invalid quality flag(s) found, only the following values are supported: {set(cls._FORWARD.keys())}"
                )

            if not causes.isin(_QUALITY_CAUSES).all(axis=None):
                raise ValueError(
                    f"invalid quality cause(s) found, only the following values are supported: {_QUALITY_CAUSES}"
                )

            if (~flags.isin(("OK", "NIL")) & (causes == "")).any(axis=None):
                raise ValueError(
                    "quality flags other than 'OK and 'NIL' need a non-empty quality cause"
                )

            if ((causes == "OTHER") & (comments == "")).any(axis=None):
                raise ValueError(
                    "quality cause 'OTHER' needs a non-empty quality comment"
                )



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

from __future__ import annotations

import functools
import inspect
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from scipy.spatial.distance import pdist
from typing_extensions import Literal

from saqc import BAD
from saqc.core import DictOfSeries, Flags, flagging, register
from saqc.funcs.changepoints import _getChangePoints
from saqc.lib.docs import DOC_TEMPLATES
from saqc.lib.tools import detectDeviants, filterKwargs, toSequence
from saqc.lib.ts_operators import expDriftModel, linearDriftModel
from saqc.lib.types import CurveFitter

if TYPE_CHECKING:
    from saqc import SaQC


LinkageString = Literal[
    "single", "complete", "average", "weighted", "centroid", "median", "ward"
]

MODELDICT = {"linear": linearDriftModel, "exponential": expDriftModel}


def cityblock(x: np.ndarray | pd.Series, y: np.ndarray | pd.Series) -> np.ndarray:
    return pdist(np.array([x, y]), metric="cityblock") / len(x)


class DriftMixin:
    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],  # reference is written !
        multivariate=True,
        handles_target=False,
        docstring=DOC_TEMPLATES["field"],
    )
    def flagDriftFromNorm(
        self: "SaQC",
        field: Sequence[str],
        freq: str,
        spread: float,
        frac: float = 0.5,
        metric: Callable[
            [np.ndarray | pd.Series, np.ndarray | pd.Series], np.ndarray
        ] = cityblock,
        method: LinkageString = "single",
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flags data that deviates from an avarage data course.

        "Normality" is determined in terms of a maximum spreading distance,
        that members of a normal group must not exceed. In addition, only a group is considered
        "normal" if it contains more then `frac` percent of the variables in "field".

        See the Notes section for a more detailed presentation of the algorithm

        Parameters
        ----------
        freq :
            Frequency, that split the data in chunks.

        spread :
            Maximum spread allowed in the group of *normal* data. See Notes section for more details.

        frac :
            Fraction defining the normal group. Use a value from the interval [0,1].
            The higher the value, the more stable the algorithm will be. For values below
            0.5 the results are undefined.

        metric : default cityblock
            Distance function that takes two arrays as input and returns a scalar float.
            This value is interpreted as the distance of the two input arrays.
            Defaults to the `averaged manhattan metric` (see Notes).

        method :
            Linkage method used for hierarchical (agglomerative) clustering of the data.
            `method` is directly passed to ``scipy.hierarchy.linkage``. See its documentation [1] for
            more details. For a general introduction on hierarchical clustering see [2].

        Notes
        -----
        following steps are performed for every data "segment" of length `freq` in order to find the
        "abnormal" data:

        1. Calculate distances :math:`d(x_i,x_j)` for all :math:`x_i` in parameter `field`.
           (with :math:`d` denoting the distance function, specified by `metric`.
        2. Calculate a dendogram with a hierarchical linkage algorithm, specified by `method`.
        3. Flatten the dendogram at the level, the agglomeration costs exceed `spread`
        4. check if a cluster containing more than `frac` variables.

            1. if yes: flag all the variables that are not in that cluster (inside the segment)
            2. if no: flag nothing

        The main parameter giving control over the algorithms behavior is the `spread` parameter,
        that determines the maximum spread of a normal group by limiting the costs, a cluster
        agglomeration must not exceed in every linkage step.
        For singleton clusters, that costs just equal half the distance, the data in the
        clusters, have to each other. So, no data can be clustered together, that are more then
        2*`spread` distances away from each other. When data get clustered together, this new
        clusters distance to all the other data/clusters is calculated according to the linkage
        method specified by `method`. By default, it is the minimum distance, the members of the
        clusters have to each other. Having that in mind, it is advisable to choose a distance
        function, that can be well interpreted in the units dimension of the measurement and where
        the interpretation is invariant over the length of the data. That is, why,
        the "averaged manhattan metric" is set as the metric default, since it corresponds to the
        averaged value distance, two data sets have (as opposed by euclidean, for example).

        References
        ----------
        Documentation of the underlying hierarchical clustering algorithm:
            [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
        Introduction to Hierarchical clustering:
            [2] https://en.wikipedia.org/wiki/Hierarchical_clustering
        """
        fields = toSequence(field)

        data = self._data[fields].to_pandas()
        data.dropna(inplace=True)

        segments = data.groupby(pd.Grouper(freq=freq))
        for segment in segments:
            if len(segment[1]) <= 1:
                continue

            drifters = detectDeviants(
                segment[1], metric, spread, frac, method, "variables"
            )

            for var in drifters:
                self._flags[segment[1].index, fields[var]] = flag

        return self

    @register(
        mask=["field", "reference"],
        demask=["field", "reference"],
        squeeze=["field", "reference"],  # reference is written !
        multivariate=True,
        handles_target=False,
    )
    def flagDriftFromReference(
        self: "SaQC",
        field: Sequence[str],
        reference: str,
        freq: str,
        thresh: float,
        metric: Callable[
            [np.ndarray | pd.Series, np.ndarray | pd.Series], np.ndarray
        ] = cityblock,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flags data that deviates from a reference course. Deviation is measured by a
        custom distance function.

        Parameters
        ----------
        freq :
            Frequency, that split the data in chunks.

        reference :
            Reference variable, the deviation is calculated from.

        thresh :
            Maximum deviation from reference.

        metric : default cityblock
            Distance function. Takes two arrays as input and returns a scalar float.
            This value is interpreted as the mutual distance of the two input arrays.
            Defaults to the `averaged manhattan metric` (see Notes).

        Notes
        -----
        It is advisable to choose a distance function, that can be well interpreted in
        the units dimension of the measurement and where the interpretation is invariant over the
        length of the data. That is, why, the "averaged manhatten metric" is set as the metric
        default, since it corresponds to the averaged value distance, two data sets have (as opposed
        by euclidean, for example).
        """

        fields = toSequence(field)

        if reference not in fields:
            fields.append(reference)

        data = self._data[fields].to_pandas().dropna()

        segments = data.groupby(pd.Grouper(freq=freq))
        for segment in segments:
            if len(segment[1]) <= 1:
                continue

            for i in range(len(fields)):
                dist = metric(
                    segment[1].iloc[:, i].values, segment[1].loc[:, reference].values
                )

                if dist > thresh:
                    self._flags[segment[1].index, fields[i]] = flag

        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def correctDrift(
        self: "SaQC",
        field: str,
        maintenance_field: str,
        model: Callable[..., float] | Literal["linear", "exponential"],
        cal_range: int = 5,
        **kwargs,
    ) -> "SaQC":
        """
        The function corrects drifting behavior.

        See the Notes section for an overview over the correction algorithm.

        Parameters
        ----------
        maintenance_field :
            Column holding the support-points information.
            The data is expected to have the following form:
            The index of the series represents the beginning of a maintenance
            event, wheras the values represent its endings.

        model :
            A model function describing the drift behavior, that is to be corrected.
            Either use built-in exponential or linear drift model by passing a string,
            or pass a custom callable. The model function must always contain the keyword
            parameters 'origin' and 'target'. The starting parameter must always be the
            parameter, by wich the data is passed to the model. After the data parameter,
            there can occure an arbitrary number of model calibration arguments in the
            signature. See the Notes section for an extensive description.

        cal_range :
            Number of values to calculate the mean of, for obtaining the value level directly
            after and directly before a maintenance event. Needed for shift calibration.

        Notes
        -----
        It is assumed, that between support points, there is a drift effect shifting the
        meassurements in a way, that can be described, by a model function M(t, p, origin, target).
        (With 0<=t<=1, p being a parameter set, and origin, target being floats).

        Note, that its possible for the model to have no free parameters p at all
        (linear drift mainly).

        The drift model, directly after the last support point (t=0),
        should evaluate to the origin - calibration level (origin), and directly before the next
        support point (t=1), it should evaluate to the target calibration level (target).


            M(0, p, origin, target) = origin
            M(1, p, origin, target) = target


        The model is than fitted to any data chunk in between support points, by optimizing
        the parameters p, and thus, obtaining optimal parameterset P.

        The new values at t are computed via:::

            new_vals(t) = old_vals(t) + M(t, P, origin, target) - M_drift(t, P, origin, new_target)

        Wheras ``new_target`` represents the value level immediately after the next support point.

        Examples
        --------
        Some examples of meaningful driftmodels.

        Linear drift modell (no free parameters).


        >>> Model = lambda t, origin, target: origin + t*target

        exponential drift model (exponential raise!)

        >>> expFunc = lambda t, a, b, c: a + b * (np.exp(c * x) - 1)
        >>> Model = lambda t, p, origin, target: expFunc(t, (target - origin) / (np.exp(abs(c)) - 1), abs(c))

        Exponential and linear driftmodels are part of the ``ts_operators`` library, under the names
        ``expDriftModel`` and ``linearDriftModel``.

        """
        # extract model func:
        if isinstance(model, str):
            if model not in MODELDICT:
                raise ValueError(
                    f"invalid model '{model}', choose one of '{MODELDICT.keys()}'"
                )
            model = MODELDICT[model]

        # 1: extract fit intervals:
        if self._data[maintenance_field].empty:
            return self

        to_correct = self._data[field].copy()
        maint_data = self._data[maintenance_field].copy()

        to_correct_clean = to_correct.dropna()
        d = {"drift_group": np.nan, field: to_correct_clean.values}
        drift_frame = pd.DataFrame(d, index=to_correct_clean.index)

        # group the drift frame
        for k in range(0, len(maint_data) - 1):
            # assign group numbers for the timespans in between one maintenance ending and the beginning of the next
            # maintenance time itself remains np.nan assigned
            drift_frame.loc[
                maint_data.values[k] : pd.Timestamp(maint_data.index[k + 1]),
                "drift_group",
            ] = k

        # define target values for correction
        drift_grouper = drift_frame.groupby("drift_group")
        shift_targets = drift_grouper.aggregate(lambda x: x[:cal_range].mean()).shift(
            -1
        )

        for k, group in drift_grouper:
            data_series = group[field]
            data_fit, data_shiftTarget = _driftFit(
                data_series, shift_targets.loc[k, :][0], cal_range, model
            )
            data_fit = pd.Series(data_fit, index=group.index)
            data_shiftTarget = pd.Series(data_shiftTarget, index=group.index)
            data_shiftVektor = data_shiftTarget - data_fit
            shiftedData = data_series + data_shiftVektor
            to_correct[shiftedData.index] = shiftedData

        self._data[field] = to_correct

        return self

    @register(mask=["field", "cluster_field"], demask=["cluster_field"], squeeze=[])
    def correctRegimeAnomaly(
        self: "SaQC",
        field: str,
        cluster_field: str,
        model: CurveFitter,
        tolerance: Optional[str] = None,
        epoch: bool = False,
        **kwargs,
    ) -> "SaQC":
        """
        Function fits the passed model to the different regimes in data[field] and tries to correct
        those values, that have assigned a negative label by data[cluster_field].

        Currently, the only correction mode supported is the "parameter propagation."

        This means, any regime :math:`z`, labeled negatively and being modeled by the parameters
        p, gets corrected via:

        :math:`z_{correct} = z + (m(p^*) - m(p))`,

        where :math:`p^*` denotes the parameter set belonging to the fit of the nearest
        not-negatively labeled cluster.

        Parameters
        ----------
        cluster_field :
            A string denoting the field in data, holding the cluster label for the data you want
            to correct.

        model :
            The model function to be fitted to the regimes.
            It must be a function of the form :math:`f(x, *p)`, where :math:`x` is the
            ``numpy.array`` holding the independent variables and :math:`p` are the model
            parameters that are to be obtained by fitting. Depending on the `x_date` parameter,
            independent variable x will either be the timestamps of every regime transformed to
            seconds from epoch, or it will be just seconds, counting the regimes length.

        tolerance :
            If an offset string is passed, a data chunk of length `offset` right at the
            start and right at the end is ignored when fitting the model. This is to account
            for the unreliability of data near the changepoints of regimes.

        epoch :
            If True, use "seconds from epoch" as x input to the model func, instead of
            "seconds from regime start".

        """
        cluster_ser = self._data[cluster_field]
        unique_successive = pd.unique(cluster_ser.values)
        data_ser = self._data[field]
        regimes = data_ser.groupby(cluster_ser)
        para_dict = {}
        x_dict = {}
        x_mask = {}
        if tolerance is not None:
            # get seconds
            tolerance = pd.Timedelta(tolerance).total_seconds()
        for label, regime in regimes:
            if epoch is False:
                # get seconds data:
                xdata = (regime.index - regime.index[0]).to_numpy(dtype=float) * 10 ** (
                    -9
                )
            else:
                # get seconds from epoch data
                xdata = regime.index.to_numpy(dtype=float) * 10 ** (-9)
            ydata = regime.values
            valid_mask = ~np.isnan(ydata)
            if tolerance is not None:
                valid_mask &= xdata > xdata[0] + tolerance
                valid_mask &= xdata < xdata[-1] - tolerance
            try:
                p, *_ = curve_fit(model, xdata[valid_mask], ydata[valid_mask])
            except (RuntimeError, ValueError):
                p = np.array([np.nan])
            para_dict[label] = p
            x_dict[label] = xdata
            x_mask[label] = valid_mask

        first_normal = unique_successive > 0
        first_valid = np.array(
            [
                ~pd.isna(para_dict[unique_successive[i]]).any()
                for i in range(0, len(unique_successive))
            ]
        )
        first_valid = np.where(first_normal & first_valid)[0][0]
        last_valid = 1

        for k in range(0, len(unique_successive)):
            if unique_successive[k] < 0 & (
                not pd.isna(para_dict[unique_successive[k]]).any()
            ):
                ydata = data_ser[regimes.groups[unique_successive[k]]].values
                xdata = x_dict[unique_successive[k]]
                ypara = para_dict[unique_successive[k]]
                if k > 0:
                    target_para = para_dict[unique_successive[k - last_valid]]
                else:
                    # first regime has no "last valid" to its left, so we use first valid to the right:
                    target_para = para_dict[unique_successive[k + first_valid]]
                y_shifted = ydata + (model(xdata, *target_para) - model(xdata, *ypara))
                data_ser[regimes.groups[unique_successive[k]]] = y_shifted
                if k > 0:
                    last_valid += 1
            elif pd.isna(para_dict[unique_successive[k]]).any() & (k > 0):
                last_valid += 1
            else:
                last_valid = 1

        self._data[field] = data_ser
        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def correctOffset(
        self: "SaQC",
        field: str,
        max_jump: float,
        spread: float,
        window: str,
        min_periods: int,
        tolerance: str | None = None,
        **kwargs,
    ) -> "SaQC":
        """
        Parameters
        ----------
        max_jump :
            when searching for changepoints in mean - this is the threshold a mean difference in the
            sliding window search must exceed to trigger changepoint detection.

        spread :
            threshold denoting the maximum, regimes are allowed to abolutely differ in their means
            to form the "normal group" of values.

        window :
            Size of the adjacent windows that are used to search for the mean changepoints.

        min_periods :
            Minimum number of periods a search window has to contain, for the result of the changepoint
            detection to be considered valid.

        tolerance :
            If an offset string is passed, a data chunk of length `offset` right from the
            start and right before the end of any regime is ignored when calculating a regimes mean for data correcture.
            This is to account for the unrelyability of data near the changepoints of regimes.

        """
        # Hint: This whole function does not set any flags

        cluster_field = field + "_CPcluster"
        self = self.copyField(field, cluster_field)
        self.data[cluster_field] = _getChangePoints(
            data=self._data[cluster_field],
            stat_func=lambda x, y: np.abs(np.mean(x) - np.mean(y)),
            thresh_func=lambda x, y: max_jump,
            window=window,
            min_periods=min_periods,
            result="cluster",
        )
        self._data, self._flags = _assignRegimeAnomaly(
            data=self._data,
            field=field,
            flags=self._flags,
            cluster_field=cluster_field,
            spread=spread,
        )
        self = self.correctRegimeAnomaly(
            field,
            cluster_field,
            lambda x, p1: np.array([p1] * x.shape[0]),
            tolerance=tolerance,
        )
        self = self.dropField(cluster_field)
        return self

    @flagging()
    def flagRegimeAnomaly(
        self: "SaQC",
        field: str,
        cluster_field: str,
        spread: float,
        method: LinkageString = "single",
        metric: Callable[
            [np.ndarray | pd.Series, np.ndarray | pd.Series], float
        ] = lambda x, y: np.abs(np.nanmean(x) - np.nanmean(y)),
        frac: float = 0.5,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flags anomalous regimes regarding to modelling regimes of ``field``.

        "Normality" is determined in terms of a maximum spreading distance,
        regimes must not exceed in respect to a certain metric and linkage method.

        In addition, only a range of regimes is considered "normal", if it models
        more then `frac` percentage of the valid samples in "field".

        Note, that you must detect the regime changepoints prior to calling this function.

        Note, that it is possible to perform hypothesis tests for regime equality
        by passing the metric a function for p-value calculation and selecting linkage
        method "complete".

        Parameters
        ----------
        cluster_field :
            Column in data, holding the cluster labels for the samples in field.
            (has to be indexed equal to field)

        spread :
            A threshold denoting the value level, up to wich clusters a agglomerated.

        method :
            The linkage method for hierarchical (agglomerative) clustering of the variables.

        metric : default absolute difference of means
            A metric function for calculating the dissimilarity between 2 regimes.
            Defaults to the difference in mean.

        frac :
            Has to be in [0,1]. Determines the minimum percentage of samples,
            the "normal" group has to comprise to be the normal group actually.
        """
        reserverd = ["set_cluster", "set_flags"]
        kwargs = filterKwargs(kwargs, reserverd)
        self._data, self._flags = _assignRegimeAnomaly(
            data=self._data,
            field=field,
            flags=self._flags,
            cluster_field=cluster_field,
            spread=spread,
            method=method,
            metric=metric,
            frac=frac,
            flag=flag,
            **kwargs,
            set_cluster=False,
            set_flags=True,
        )
        return self

    @register(mask=["field", "cluster_field"], demask=["cluster_field"], squeeze=[])
    def assignRegimeAnomaly(
        self: "SaQC",
        field: str,
        cluster_field: str,
        spread: float,
        method: LinkageString = "single",
        metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: np.abs(
            np.nanmean(x) - np.nanmean(y)
        ),
        frac: float = 0.5,
        **kwargs,
    ) -> "SaQC":
        """
        A function to detect values belonging to an anomalous regime regarding modelling
        regimes of field.

        The function changes the value of the regime cluster labels to be negative.
        "Normality" is determined in terms of a maximum spreading distance, regimes must
        not exceed in respect to a certain metric and linkage method. In addition,
        only a range of regimes is considered "normal", if it models more then `frac`
        percentage of the valid samples in "field". Note, that you must detect the regime
        changepoints prior to calling this function. (They are expected to be stored
        parameter `cluster_field`.)

        Note, that it is possible to perform hypothesis tests for regime equality by
        passing the metric a function for p-value calculation and selecting linkage
        method "complete".

        Parameters
        ----------
        cluster_field :
            Column in data, holding the cluster labels for the samples in field.
            (has to be indexed equal to field)

        spread :
            A threshold denoting the value level, up to wich clusters a agglomerated.

        method :
            The linkage method for hierarchical (agglomerative) clustering of the variables.

        metric : default absolute difference of means
            A metric function for calculating the dissimilarity between 2 regimes.
            Defaults to the difference in mean.

        frac :
            Has to be in [0,1]. Determines the minimum percentage of samples,
            the "normal" group has to comprise to be the normal group actually.
        """
        reserverd = ["set_cluster", "set_flags", "flag"]
        kwargs = filterKwargs(kwargs, reserverd)
        self._data, self._flags = _assignRegimeAnomaly(
            data=self._data,
            field=field,
            flags=self._flags,
            cluster_field=cluster_field,
            spread=spread,
            method=method,
            metric=metric,
            frac=frac,
            **kwargs,
            # control args
            set_cluster=True,
            set_flags=False,
        )
        return self


def _driftFit(x, shift_target, cal_mean, driftModel):
    x_index = x.index - x.index[0]
    x_data = x_index.total_seconds().values
    x_data = x_data / x_data[-1] if len(x_data) > 1 else x_data
    y_data = x.values
    origin_mean = np.mean(y_data[:cal_mean])
    target_mean = np.mean(y_data[-cal_mean:])

    dataFitFunc = functools.partial(driftModel, origin=origin_mean, target=target_mean)
    # if drift model has free parameters:
    if len(inspect.getfullargspec(dataFitFunc).args) > 1:
        try:
            # try fitting free parameters
            fit_paras, *_ = curve_fit(dataFitFunc, x_data, y_data)
            data_fit = dataFitFunc(x_data, *fit_paras)
            data_shift = driftModel(
                x_data, *fit_paras, origin=origin_mean, target=shift_target
            )
        except RuntimeError:
            # if fit fails -> make no correction
            data_fit = np.array([0] * len(x_data))
            data_shift = np.array([0] * len(x_data))
    # when there are no free parameters in the model:
    else:
        data_fit = dataFitFunc(x_data)
        data_shift = driftModel(x_data, origin=origin_mean, target=shift_target)

    return data_fit, data_shift


def _assignRegimeAnomaly(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    cluster_field: str,
    spread: float,
    method: LinkageString = "single",
    metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: np.abs(
        np.nanmean(x) - np.nanmean(y)
    ),
    frac: float = 0.5,
    set_cluster: bool = True,
    set_flags: bool = False,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    series = data[cluster_field]
    cluster = np.unique(series)
    cluster_dios = DictOfSeries({str(i): data[field][series == i] for i in cluster})
    plateaus = detectDeviants(cluster_dios, metric, spread, frac, method, "samples")

    if set_flags:
        for p, cols in zip(plateaus, cluster_dios.columns[plateaus]):
            flags[cluster_dios[cols].index, field] = flag

    if set_cluster:
        for p in plateaus:
            if cluster[p] > 0:
                series[series == cluster[p]] = -cluster[p]

    data[cluster_field] = series
    return data, flags



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from saqc.core.translation.basescheme import (
    FloatScheme,
    MappingScheme,
    TranslationScheme,
)
from saqc.core.translation.dmpscheme import DmpScheme
from saqc.core.translation.positionalscheme import PositionalScheme
from saqc.core.translation.simplescheme import SimpleScheme



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import numpy as np
import pytest

import saqc.lib.ts_operators as ts_ops
from saqc.core import DictOfSeries, Flags, SaQC, flagging
from saqc.parsing.environ import ENVIRONMENT
from saqc.parsing.reader import fromConfig, readFile
from tests.common import initData, writeIO


@pytest.fixture
def data() -> DictOfSeries:
    return initData(3)


def getTestedVariables(flags: Flags, test: str):
    out = []
    for col in flags.columns:
        for m in flags.history[col].meta:
            if m["func"] == test:
                out.append(col)
    return out


def test_variableRegex(data):
    header = f"varname;test"
    function = "flagDummy"
    tests = [
        ("'.*'", data.columns),
        ("'var(1|2)'", [c for c in data.columns if c[-1] in ("1", "2")]),
        ("'var[12]'", [c for c in data.columns if c[-1] in ("1", "2")]),
        ('".*3"', [c for c in data.columns if c[-1] == "3"]),
    ]

    for regex, expected in tests:
        fobj = writeIO(header + "\n" + f"{regex} ; {function}()")
        saqc = fromConfig(fobj, data=data)
        result = getTestedVariables(saqc._flags, function)
        assert np.all(result == expected)

    tests = [
        ("var[12]", []),  # not quoted -> not a regex
    ]
    for regex, expected in tests:
        fobj = writeIO(header + "\n" + f"{regex} ; {function}()")
        with pytest.warns(RuntimeWarning):
            saqc = fromConfig(fobj, data=data)
        result = getTestedVariables(saqc._flags, function)
        assert np.all(result == expected)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_inlineComments(data):
    """
    adresses issue #3
    """
    config = f"""
    varname ; test
    var1    ; flagDummy() # test
    """

    saqc = fromConfig(writeIO(config), data)
    func = saqc._flags.history["var1"].meta[0]["func"]
    assert func == "flagDummy"


def test_configReaderLineNumbers():
    config = f"""
    varname ; test
    #temp1      ; flagDummy()
    pre1        ; flagDummy()
    pre2        ; flagDummy()
    SM          ; flagDummy()
    #SM         ; flagDummy()
    # SM1       ; flagDummy()

    SM1         ; flagDummy()
    """
    planned = readFile(writeIO(config))
    expected = [4, 5, 6, 10]
    assert (planned.index == expected).all()


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_configFile(data):
    # check that the reader accepts different whitespace patterns

    config = f"""
    varname ; test

    #temp1      ; flagDummy()
    pre1; flagDummy()
    pre2        ;flagDummy()
    SM          ; flagDummy()
    #SM         ; flagDummy()
    # SM1       ; flagDummy()

    SM1;flagDummy()
    """
    fromConfig(writeIO(config), data)


def test_configChecks(data):
    var1, _, var3, *_ = data.columns

    @flagging()
    def flagFunc(data, field, flags, arg, opt_arg=None, **kwargs):
        flags[:, field] = np.nan
        return data, flags

    header = f"varname;test"
    tests = [
        (f"{var1};flagFunc(mn=0)", TypeError),  # bad argument name
        (f"{var1};flagFunc()", TypeError),  # not enough arguments
        (f"{var3};flagNothing()", NameError),  # unknown function
        (f"{var1}; min", TypeError),  # not a function call
    ]

    for test, expected in tests:
        fobj = writeIO(header + "\n" + test)
        with pytest.raises(expected):
            fromConfig(fobj, data=data)


def test_supportedArguments(data):
    # test if the following function arguments
    # are supported (i.e. parsing does not fail)

    # TODO: necessary?

    @flagging()
    def func(saqc, field, kwarg, **kwargs):
        saqc._flags[:, field] = np.nan
        return saqc

    var1 = data.columns[0]

    header = f"varname;test"
    tests = [
        f"{var1};func(kwarg=NAN)",
        f"{var1};func(kwarg='str')",
        f"{var1};func(kwarg=5)",
        f"{var1};func(kwarg=5.5)",
        f"{var1};func(kwarg=-5)",
        f"{var1};func(kwarg=True)",
        f"{var1};func(kwarg=sum([1, 2, 3]))",
    ]

    for test in tests:
        fobj = writeIO(header + "\n" + test)
        fromConfig(fobj, data)


@pytest.mark.parametrize(
    "func_string", [k for k, v in ENVIRONMENT.items() if callable(v)]
)
def test_funtionArguments(data, func_string):
    @flagging()
    def testFunction(saqc, field, func, **kwargs):
        assert func is ENVIRONMENT[func_string]
        return saqc

    config = f"""
    varname ; test
    {data.columns[0]} ; testFunction(func={func_string})
    {data.columns[0]} ; testFunction(func="{func_string}")
    """

    fromConfig(writeIO(config), data)



# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any, TypedDict

from docstring_parser import DocstringParam, DocstringReturns, compose, parse


class ParamDict(TypedDict):
    typehint: str | None
    description: str | None
    optional: bool | None


DOC_TEMPLATES = {
    "field": {
        "typehint": "List[str]",
        "description": "List of variables names to process.",
    },
    "target": {"optional": False},
}

COMMON = {
    "field": {
        "name": "field",
        "description": "Variable to process.",
        "typehint": "str | list[str]",
    },
    "target": {
        "name": "target",
        "description": "Variable name to which the results are written. :py:attr:`target` will be created if it does not exist. Defaults to :py:attr:`field`.",
        "typehint": "str | list[str]",
        "optional": True,
    },
    "dfilter": {
        "name": "dfilter",
        "description": "Defines which observations will be masked based on the already existing flags. Any data point with a flag equal or worse to this threshold will be passed as ``NaN`` to the function. Defaults to the ``DFILTER_ALL`` value of the translation scheme.",
        "typehint": "Any",
        "optional": True,
    },
    "flag": {
        "name": "flag",
        "description": "The flag value the function uses to mark observations. Defaults to the ``BAD`` value of the translation scheme.",
        "typehint": "Any",
        "optional": True,
    },
}


class FunctionParam(DocstringParam):
    def __init__(
        self, name: str, typehint: str, description: str, optional: bool = False
    ):
        super().__init__(
            args=["param", name],
            description=description,
            arg_name=name,
            type_name=typehint,
            is_optional=optional,
            default=None,
        )


def docurator(func, defaults: dict[str, ParamDict] | None = None):
    if defaults is None:
        defaults = {}

    docstring_return = DocstringReturns(
        args=["returns"],
        description="the updated SaQC object",
        type_name="saqc.SaQC",
        is_generator=False,
        return_name="SaQC",
    )

    tree = parse(func.__doc__)

    if tree.returns:
        raise ValueError(
            f"'{func.__name__}' function doctstring should not provide a returns section"
        )

    # rewrite parameters
    meta = [FunctionParam(**{**COMMON["field"], **defaults.get("field", {})})]
    for p in tree.params:
        if p.arg_name in COMMON:
            raise ValueError(
                f"'{func.__name__}' function docstring should not provide a parameter description for '{p.arg_name}'"
            )
        meta.append(p)

    # additional parameters
    for p in ("target", "dfilter", "flag"):
        meta.append(FunctionParam(**{**COMMON[p], **defaults.get(p, {})}))

    # return sections
    meta.append(docstring_return)

    # everyhing else the docstring provides
    for m in tree.meta:
        if not isinstance(m, DocstringParam):
            meta.append(m)

    tree.meta = meta

    func.__doc__ = compose(tree)
    return func



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

import numpy as np
import pandas as pd

from saqc.constants import BAD, DOUBTFUL, GOOD, UNFLAGGED
from saqc.core import Flags, History
from saqc.core.translation.basescheme import BackwardMap, ForwardMap, MappingScheme


class PositionalScheme(MappingScheme):

    """
    Implements the translation from and to the flagging scheme implemented by CHS
    """

    DFILTER_DEFAULT = DOUBTFUL + 1

    _FORWARD: ForwardMap = {
        -6: UNFLAGGED,
        -5: UNFLAGGED,
        -2: UNFLAGGED,
        -1: UNFLAGGED,
        0: UNFLAGGED,
        1: DOUBTFUL,
        2: BAD,
    }
    _BACKWARD: BackwardMap = {
        np.nan: 0,
        UNFLAGGED: 0,
        GOOD: 0,
        DOUBTFUL: 1,
        BAD: 2,
    }

    def __init__(self):
        super().__init__(forward=self._FORWARD, backward=self._BACKWARD)

    def toInternal(self, flags: pd.DataFrame) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """

        data = {}
        for field, field_flags in flags.items():
            # explode the flags into sperate columns and drop the leading `9`
            df = pd.DataFrame(
                field_flags.astype(str).str.slice(start=1).apply(tuple).tolist(),
                index=field_flags.index,
            ).astype(int)

            # the exploded values form the History of `field`
            fflags = super()._translate(df, self._FORWARD)
            field_history = History(field_flags.index)
            for _, s in fflags.items():
                field_history.append(s)
            data[str(field)] = field_history

        return Flags(data)

    def toExternal(self, flags: Flags, **kwargs) -> pd.DataFrame:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        pd.DataFrame
        """
        out = {}
        for field in flags.columns:
            thist = flags.history[field].hist.replace(self._BACKWARD).astype(float)
            # concatenate the single flag values
            ncols = len(thist.columns)
            init = 9 * 10**ncols
            bases = 10 ** np.arange(ncols - 1, -1, -1)

            tflags = init + (thist * bases).sum(axis=1)
            out[field] = tflags

        return pd.DataFrame(out).fillna(-9999).astype(int)



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING

import dtw
import numpy as np
import pandas as pd

from saqc import BAD
from saqc.core import flagging
from saqc.lib.rolling import removeRollingRamps

if TYPE_CHECKING:
    from saqc import SaQC


def calculateDistanceByDTW(
    data: pd.Series, reference: pd.Series, forward=True, normalize=True
):
    """
    Calculate the DTW-distance of data to pattern in a rolling calculation.

    The data is compared to pattern in a rolling window.
    The size of the rolling window is determined by the timespan defined
    by the first and last timestamp of the reference data's datetime index.

    For details see the linked functions in the `See Also` section.

    Parameters
    ----------
    data : pd.Series
        Data series. Must have datetime-like index, and must be regularly sampled.

    reference : : pd.Series
        Reference series. Must have datetime-like index, must not contain NaNs
        and must not be empty.

    forward: bool, default True
        If `True`, the distance value is set on the left edge of the data chunk. This
        means, with a perfect match, `0.0` marks the beginning of the pattern in
        the data. If `False`, `0.0` would mark the end of the pattern.

    normalize : bool, default True
        If `False`, return unmodified distances.
        If `True`, normalize distances by the number of observations in the reference.
        This helps to make it easier to find a good cutoff threshold for further
        processing. The distances then refer to the mean distance per datapoint,
        expressed in the datas units.

    Returns
    -------
    distance : pd.Series

    Notes
    -----
    The data must be regularly sampled, otherwise a ValueError is raised.
    NaNs in the data will be dropped before dtw distance calculation.

    See Also
    --------
    flagPatternByDTW : flag data by DTW
    """
    if reference.hasnans or reference.empty:
        raise ValueError("reference must not have nan's and must not be empty.")

    winsz: pd.Timedelta = reference.index.max() - reference.index.min()
    reference = reference.to_numpy()

    def isPattern(chunk):
        if forward:
            return dtw.accelerated_dtw(chunk[::-1], reference, "euclidean")[0]
        else:
            return dtw.accelerated_dtw(chunk, reference, "euclidean")[0]

    # generate distances, excluding NaNs
    nonas = data.dropna()
    rollover = nonas[::-1] if forward else nonas
    arr = rollover.rolling(winsz, closed="both").apply(isPattern, raw=True).to_numpy()
    distances = pd.Series(arr[::-1] if forward else arr, index=nonas.index)
    removeRollingRamps(distances, window=winsz, inplace=True)

    if normalize:
        distances /= len(reference)

    return distances.reindex(index=data.index)  # reinsert NaNs


class PatternMixin:
    # todo should we mask `reference` even if the func fail if reference has NaNs
    @flagging()
    def flagPatternByDTW(
        self: "SaQC",
        field,
        reference,
        max_distance=0.0,
        normalize=True,
        plot=False,
        flag=BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Pattern Recognition via Dynamic Time Warping.

        The steps are:
        1. work on a moving window

        2. for each data chunk extracted from each window, a distance to the given pattern
           is calculated, by the dynamic time warping algorithm [1]

        3. if the distance is below the threshold, all the data in the window gets flagged

        Parameters
        ----------
        reference :
            The name in `data` which holds the pattern. The pattern must not have NaNs,
            have a datetime index and must not be empty.

        max_distance :
            Maximum dtw-distance between chunk and pattern, if the distance is lower than
            ``max_distance`` the data gets flagged. With default, ``0.0``, only exact
            matches are flagged.

        normalize :
            If `False`, return unmodified distances.
            If `True`, normalize distances by the number of observations of the reference.
            This helps to make it easier to find a good cutoff threshold for further
            processing. The distances then refer to the mean distance per datapoint,
            expressed in the datas units.

        plot :
            Show a calibration plot, which can be quite helpful to find the right threshold
            for `max_distance`. It works best with `normalize=True`. Do not use in automatic
            setups / pipelines. The plot show three lines:

            - data: the data the function was called on
            - distances: the calculated distances by the algorithm
            - indicator: have to distinct levels: `0` and the value of `max_distance`.
              If `max_distance` is `0.0` it defaults to `1`. Everywhere where the
              indicator is not `0` the data will be flagged.

        Notes
        -----
        The window size of the moving window is set to equal the temporal extension of the
        reference datas datetime index.

        References
        ----------
        Find a nice description of underlying the Dynamic Time Warping Algorithm here:

        [1] https://cran.r-project.org/web/packages/dtw/dtw.pdf
        """
        ref = self._data[reference]
        dat = self._data[field]

        distances = calculateDistanceByDTW(dat, ref, forward=True, normalize=normalize)
        winsz = ref.index.max() - ref.index.min()

        # prevent nan propagation
        distances = distances.fillna(max_distance + 1)

        # find minima filter by threshold
        fw_min = distances[::-1].rolling(window=winsz, closed="both").min()[::-1]
        bw_min = distances.rolling(window=winsz, closed="both").min()
        minima = (fw_min == bw_min) & (distances <= max_distance)

        # Propagate True's to size of pattern.
        mask = minima.rolling(window=winsz, closed="both").sum() > 0

        if plot:
            df = pd.DataFrame()
            df["data"] = dat
            df["distances"] = distances
            df["indicator"] = mask.astype(float) * (max_distance or 1)
            df.plot()

        self._flags[mask, field] = flag
        return self



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional, Union

import numpy as np
import pandas as pd

from saqc.core import register
from saqc.funcs.curvefit import _fitPolynomial
from saqc.funcs.rolling import _roll

if TYPE_CHECKING:
    from saqc import SaQC


class ResidualsMixin:
    @register(mask=["field"], demask=[], squeeze=[])
    def calculatePolynomialResiduals(
        self: "SaQC",
        field: str,
        window: str | int,
        order: int,
        min_periods: int = 0,
        **kwargs,
    ) -> "SaQC":
        """
        Fits a polynomial model to the data and calculate the residuals.

        The residual  is calculated by fitting a polynomial of degree `order` to a data
        slice of size `window`, that has x at its center.

        Note, that calculating the residuals tends to be quite costy, because a function
        fitting is performed for every sample. To improve performance, consider the
        following possibilities:

        In case your data is sampled at an equidistant frequency grid:

        (1) If you know your data to have no significant number of missing values,
        or if you do not want to calculate residuals for windows containing missing values
        any way, performance can be increased by setting min_periods=window.

        Note, that the initial and final window/2 values do not get fitted.

        Each residual gets assigned the worst flag present in the interval of
        the original data.

        Parameters
        ----------
        window :
            The size of the window you want to use for fitting. If an integer is passed,
            the size refers to the number of periods for every fitting window. If an
            offset string is passed, the size refers to the total temporal extension. The
            window will be centered around the vaule-to-be-fitted. For regularly sampled
            timeseries the period number will be casted down to an odd number if even.

        order :
            The degree of the polynomial used for fitting

        min_periods :
            The minimum number of periods, that has to be available in every values
            fitting surrounding for the polynomial fit to be performed. If there are not
            enough values, np.nan gets assigned. Default (0) results in fitting
            regardless of the number of values present (results in overfitting for too
            sparse intervals). To automatically set the minimum number of periods to the
            number of values in an offset defined window size, pass np.nan.
        """
        orig = self._data[field]
        data, _ = _fitPolynomial(
            data=self._data,
            field=field,
            flags=self._flags,
            window=window,
            order=order,
            min_periods=min_periods,
            **kwargs,
        )
        self._data[field] = orig - data[field]
        return self

    @register(mask=["field"], demask=[], squeeze=[])
    def calculateRollingResiduals(
        self: "SaQC",
        field: str,
        window: str | int,
        func: Callable[[pd.Series], np.ndarray] = np.mean,
        min_periods: int = 0,
        center: bool = True,
        **kwargs,
    ) -> "SaQC":
        """
        Calculate the diff of a rolling-window function and the data.

        Note, that the data gets assigned the worst flag present in the original data.

        Parameters
        ----------
        window :
            The size of the window you want to roll with. If an integer is passed, the size
            refers to the number of periods for every fitting window. If an offset string
            is passed, the size refers to the total temporal extension. For regularly
            sampled timeseries, the period number will be casted down to an odd number if
            ``center=True``.

        func : default mean
            Function to roll with.

        min_periods :
            The minimum number of periods to get a valid value

        center :
            If True, center the rolling window.
        """
        orig = self._data[field].copy()
        data, _ = _roll(
            data=self._data,
            field=field,
            flags=self._flags,
            window=window,
            func=func,
            min_periods=min_periods,
            center=center,
            **kwargs,
        )

        # calculate residual
        self._data[field] = orig - data[field]
        return self



#! /usr/bin/env python
# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
# SPDX-License-Identifier: GPL-3.0-or-later
# -*- coding: utf-8 -*-

import numpy as np

from saqc.constants import BAD, GOOD, UNFLAGGED
from saqc.core.translation import MappingScheme


class SimpleScheme(MappingScheme):

    """
    Acts as the default Translator, provides a changeable subset of the
    internal float flags
    """

    _FORWARD = {
        "UNFLAGGED": UNFLAGGED,
        "BAD": BAD,
        "OK": GOOD,
    }

    _BACKWARD = {
        UNFLAGGED: "UNFLAGGED",
        np.nan: "UNFLAGGED",
        BAD: "BAD",
        GOOD: "OK",
    }

    def __init__(self):
        super().__init__(forward=self._FORWARD, backward=self._BACKWARD)



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

from __future__ import annotations

import uuid
import warnings
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Tuple

import numba
import numpy as np
import numpy.polynomial.polynomial as poly
import pandas as pd
from outliers import smirnov_grubbs  # noqa, on pypi as outlier-utils
from scipy.stats import median_abs_deviation
from typing_extensions import Literal

from saqc import BAD, UNFLAGGED
from saqc.core import DictOfSeries, Flags, flagging, register
from saqc.funcs.scores import _univarScoring
from saqc.lib.docs import DOC_TEMPLATES
from saqc.lib.tools import getFreqDelta, isflagged, toSequence

if TYPE_CHECKING:
    from saqc import SaQC


class OutliersMixin:
    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],
        multivariate=True,
        handles_target=False,
    )
    def flagLOF(
        self: "SaQC",
        field: Sequence[str],
        n: int = 20,
        thresh: Literal["auto"] | float = 1.5,
        algorithm: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
        p: int = 1,
        density: Literal["auto"] | float | Callable = "auto",
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag values where the Local Outlier Factor (LOF) exceeds cutoff.

        Parameters
        ----------
        n :
            Number of neighbors to be included into the LOF calculation. Defaults to ``20``, which is a
            value found to be suitable in the literature.

            * :py:attr:`n` determines the "locality" of an observation (its :py:attr:`n` nearest neighbors)
              and sets the upper limit to the number of values in outlier clusters (i.e. consecutive outliers). Outlier
              clusters of size greater than :py:attr:`n`/2 may not be detected reliably.
            * The larger :py:attr:`n`, the lesser the algorithm's sensitivity to local outliers and small
              or singleton outliers points. Higher values greatly increase numerical costs.

        thresh :
            The threshold for flagging the calculated LOF. A LOF of around ``1`` is considered normal and
            most likely corresponds to inlier points.

            * The "automatic" threshing introduced with the publication of the algorithm defaults to ``1.5``.
            * In this implementation, :py:attr:`thresh` defaults (``'auto'``) to flagging the scores with a
              modified 3-sigma rule, resulting in a :py:attr:`thresh` `` > 1.5`` which usually mitigates
              overflagging compared to the literature recommendation.

        algorithm :
            Algorithm used for calculating the :py:attr:`n`-nearest neighbors.

        p :
            Degree of the metric ("Minkowski"), according to which the distance to neighbors is determined.
            Most important values are:

            * ``1`` - Manhatten Metric
            * ``2`` - Euclidian Metric

        Notes
        -----
        * The :py:meth:`~saqc.SaQC.flagLOF` function calculates the Local Outlier Factor (LOF) for every point
          in the input timeseries. The *LOF* is a scalar value, that roughly correlates to the *reachability*,
          or "outlierishnes" of the evaluated datapoint. If a point is as reachable, as all its
          :py:attr:`n`-nearest neighbors, the *LOF* score evaluates to around ``1``. If it is only as half as
          reachable as all its ``n``-nearest neighbors are (so to say, as double as "outlierish"), the score
          is about ``2``. So, the Local Outlier *Factor* relates a point's *reachability* to the *reachability*
          of its :py:attr:`n`-nearest neighbors in a multiplicative fashion (as a "factor").
        * The *reachability* of a point thereby is determined as an aggregation of the points distances to its
          :py:attr:`n`-nearest neighbors, measured with regard to the minkowski metric of degree :py:attr:`p`
          (usually euclidean).
        * To derive a binary label for every point (outlier: *yes*, or *no*), the scores are cut off at a level,
          determined by :py:attr:`thresh`.

        """
        fields = toSequence(field)
        field_ = str(uuid.uuid4())
        self = self.assignLOF(
            field=fields,
            target=field_,
            n=n,
            algorithm=algorithm,
            p=p,
            density=density,
        )
        s = self.data[field_]
        if thresh == "auto":
            s = pd.concat([s, (-s - 2)])
            s_mask = (s - s.mean() / s.std())[: len(s) // 2].abs() > 3
        else:
            s_mask = s < abs(thresh)

        for f in fields:
            mask = ~isflagged(self._flags[f], kwargs["dfilter"]) & s_mask
            self._flags[mask, f] = flag

        return self.dropField(field_)

    @flagging()
    def flagUniLOF(
        self: "SaQC",
        field: str,
        n: int = 20,
        thresh: Literal["auto"] | float = 1.5,
        algorithm: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
        p: int = 1,
        density: Literal["auto"] | float | Callable = "auto",
        fill_na: str = "linear",
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag "univariate" Local Outlier Factor (LOF) exceeding cutoff.

        The function is a wrapper around a usual LOF implementation, aiming for an easy to use,
        parameter minimal outlier detection function for single variables, that does not necessitate
        prior modelling of the variable. LOF is applied onto a concatenation of the `field` variable
        and a "temporal density", or "penalty" variable, that measures temporal distance between data
        points. See notes Section for a more exhaustive explaination.

        See the Notes section for more details on the algorithm.

        Parameters
        ----------
        n :
            Number of periods to be included into the LOF calculation. Defaults to `20`, which is a
            value found to be suitable in the literature.

            * :py:attr:`n` determines the "locality" of an observation (its :py:attr:`n` nearest neighbors)
              and sets the upper limit to the number of values in an outlier clusters (i.e. consecutive outliers). Outlier
              clusters of size greater than :py:attr:`n`/2 may not be detected reliably.
            * The larger :py:attr:`n`, the lesser the algorithm's sensitivity to local outliers and small
              or singleton outlier points. Higher values greatly increase numerical costs.

        thresh :
            The threshold for flagging the calculated LOF. A LOF of around ``1`` is considered normal and
            most likely corresponds to inlier points. This parameter is considered the main calibration
            parameter of the algorithm.

            * The threshing defaults to ``1.5``, wich is the default value found to be suitable in the literature.
            * ``'auto'`` enables flagging the scores with a modified 3-sigma rule,
              resulting in a thresh around ``4``, which usually greatly mitigates overflagging compared to the
              literature recommendation, but often is too high.
            * sensitive range for the parameter may be ``[1,15]``, assuming default settings for the other parameters.

        algorithm :
            Algorithm used for calculating the :py:attr:`n`-nearest neighbors needed for LOF calculation.
        p :
            Degree of the metric ("Minkowski"), according to which distance to neighbors is determined.
            Most important values are:
            * ``1`` - Manhatten Metric
            * ``2`` - Euclidian Metric
        density :
            How to calculate the temporal distance/density for the variable to flag.

            * ``'auto'`` - introduces linear density with an increment equal to the median of the absolute
              diff of the variable to flag.
            * ``float`` - introduces linear density with an increment equal to :py:attr:`density`
            * Callable - calculates the density by applying the function passed onto the variable to flag
              (passed as Series).

        fill_na :
            Weather or not to fill NaN values in the data with a linear interpolation.

        See Also
        --------
        :ref:`introduction to outlier detection with saqc <cookbooks/OutlierDetection:Outlier Detection>`

        Notes
        -----
        * The :py:meth:`~saqc.SaQC.flagUniLOF` function calculates an univariate
          Local Outlier Factor (UniLOF) - score for every point in the one dimensional input
          data series.
          The *UniLOF* score of any data point is a scalar value, that roughly correlates to
          its *reachability*, or "outlierishnes" in the 2-dimensional space constituted by the
          data-values and the time axis. So the Algorithm basically operates on the "graph",
          or the "plot" of the input timeseries.
        * If a point in this "graph" is as reachable, as all its :py:attr:`n`-nearest
          neighbors, its *UniLOF* score evaluates to around ``1``. If it is only as half as
          reachable as all its :py:attr:`n` neighbors are
          (so to say, as double as "outlierish"), its score evaluates to ``2`` roughly.
          So, the Univariate Local Outlier *Factor* relates a points *reachability* to the
          *reachability* of its :py:attr:`n`-nearest neighbors in a multiplicative fashion
          (as a "factor").
        * The *reachability* of a point thereby is derived as an aggregation of the points
          distance to its :py:attr:`n`-nearest neighbors, measured with regard to the minkowski
          metric of degree :py:attr:`p` (usually euclidean).
        * The parameter :py:attr:`density` thereby determines how dimensionality of the time is
          removed, to make it a dimensionless, real valued coordinate.
        * To derive a binary label for every point (outlier: *yes*, or *no*), the scores are cut
          off at a level, determined by :py:attr:`thresh`.

        Examples
        --------

        See the :ref:`outlier detection cookbook <cookbooks/OutlierDetection:Outlier Detection>` for a detailed
        introduction into the usage and tuning of the function.

        .. plot::
           :context: reset
           :include-source: False

           import matplotlib
           import saqc
           import pandas as pd
           data = pd.read_csv('../resources/data/hydro_data.csv')
           data = data.set_index('Timestamp')
           data.index = pd.DatetimeIndex(data.index)
           qc = saqc.SaQC(data)

        Example usage with default parameter configuration:

        Loading data via pandas csv file parser, casting index to DateTime, generating a :py:class:`~saqc.SaQC`
        instance from the data and plotting the variable representing light scattering at 254 nanometers wavelength.

        .. doctest:: flagUniLOFExample

           >>> import saqc
           >>> data = pd.read_csv('./resources/data/hydro_data.csv')
           >>> data = data.set_index('Timestamp')
           >>> data.index = pd.DatetimeIndex(data.index)
           >>> qc = saqc.SaQC(data)
           >>> qc.plot('sac254_raw') # doctest: +SKIP

        .. plot::
           :context:
           :include-source: False
           :class: center

            qc.plot('sac254_raw')

        We apply :py:meth:`~saqc.SaqC.flagUniLOF` in with default parameter values. Meaning, that the main
        calibration paramters :py:attr:`n` and :py:attr:`thresh` evaluate to `20` and `1.5` respectively.

        .. doctest:: flagUniLOFExample

           >>> import saqc
           >>> qc = qc.flagUniLOF('sac254_raw')
           >>> qc.plot('sac254_raw') # doctest: +SKIP

        .. plot::
           :context: close-figs
           :include-source: False
           :class: center

           qc = qc.flagUniLOF('sac254_raw')
           qc.plot('sac254_raw')

        """
        field_ = str(uuid.uuid4())
        self = self.assignUniLOF(
            field=field,
            target=field_,
            n=n,
            algorithm=algorithm,
            p=p,
            density=density,
            fill_na=fill_na,
        )
        s = self.data[field_]
        if thresh == "auto":
            _s = pd.concat([s, (-s - 2)])
            s_mask = ((_s - _s.mean()) / _s.std()).iloc[: int(s.shape[0])].abs() > 3
        else:
            s_mask = s < -abs(thresh)

        s_mask = ~isflagged(self._flags[field], kwargs["dfilter"]) & s_mask
        self._flags[s_mask, field] = flag
        self = self.dropField(field_)
        return self

    @flagging()
    def flagRange(
        self: "SaQC",
        field: str,
        min: float = -np.inf,
        max: float = np.inf,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Function flags values exceeding the closed interval [:py:attr:`min`, :py:attr:`max`].

        Parameters
        ----------
        min :
            Lower bound for valid data.
        max :
            Upper bound for valid data.
        """

        # using .values is much faster
        datacol = self._data[field].to_numpy()
        mask = (datacol < min) | (datacol > max)
        self._flags[mask, field] = flag
        return self

    @flagging()
    def flagByStray(
        self: "SaQC",
        field: str,
        window: int | str | None = None,
        min_periods: int = 11,
        iter_start: float = 0.5,
        alpha: float = 0.05,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag outliers in 1-dimensional (score) data using the STRAY Algorithm.

        For more details about the algorithm please refer to [1].

        Parameters
        ----------

        window :
            Determines the segmentation of the data into partitions, the kNN algorithm is
            applied onto individually.

            * ``None``: Apply Scoring on whole data set at once
            * ``int``: Apply scoring on successive data chunks of periods with the given length.
              Must be greater than 0.
            * Offset String : Apply scoring on successive partitions of temporal extension
              matching the passed offset string

        min_periods :
            Minimum number of periods per partition that have to be present for a valid
            outlier detection to be made in this partition (only of effect, if :py:attr:`freq`
            is an integer).

        iter_start :
            Float in ``[0, 1]`` that determines which percentage of data is considered
            "normal". ``0.5`` results in the stray algorithm to search only the upper 50% of
            the scores for the cut off point. (See reference section for more information)

        alpha :
            Level of significance by which it is tested, if a score might be drawn from
            another distribution than the majority of the data.

        References
        ----------
        [1]  Priyanga Dilini Talagala, Rob J. Hyndman & Kate Smith-Miles (2021):
             Anomaly Detection in High-Dimensional Data,
             Journal of Computational and Graphical Statistics, 30:2, 360-374,
             DOI: 10.1080/10618600.2020.1807997
        """
        scores = self._data[field].dropna()

        if scores.empty:
            return self

        if not window:
            window = len(scores)

        if isinstance(window, str):
            partitions = scores.groupby(pd.Grouper(freq=window))

        else:
            grouper_series = pd.Series(
                data=np.arange(0, len(scores)), index=scores.index
            )
            grouper_series = grouper_series.transform(
                lambda x: int(np.floor(x / window))
            )
            partitions = scores.groupby(grouper_series)

        # calculate flags for every window
        for _, partition in partitions:
            if partition.empty | (len(partition) < min_periods):
                continue

            sample_size = len(partition)

            sorted_i = partition.values.argsort()
            resids = partition.values[sorted_i]
            gaps = np.append(0, np.diff(resids))

            tail_size = int(max(min(50, np.floor(sample_size / 4)), 2))
            tail_indices = np.arange(2, tail_size + 1)

            i_start = int(max(np.floor(sample_size * iter_start), 1) + 1)
            ghat = np.array([np.nan] * sample_size)

            for i in range(i_start - 1, sample_size):
                ghat[i] = sum(
                    (tail_indices / (tail_size - 1)) * gaps[i - tail_indices + 1]
                )

            log_alpha = np.log(1 / alpha)
            for iter_index in range(i_start - 1, sample_size):
                if gaps[iter_index] > log_alpha * ghat[iter_index]:
                    index = partition.index[sorted_i[iter_index:]]
                    self._flags[index, field] = flag
                    break

        return self

    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],
        multivariate=True,
        handles_target=False,
        docstring={"field": DOC_TEMPLATES["field"]},
    )
    def flagMVScores(
        self: "SaQC",
        field: Sequence[str],
        trafo: Callable[[pd.Series], pd.Series] = lambda x: x,
        alpha: float = 0.05,
        n: int = 10,
        func: Callable[[pd.Series], float] = np.sum,
        iter_start: float = 0.5,
        window: int | str | None = None,
        min_periods: int = 11,
        stray_range: str | None = None,
        drop_flagged: bool = False,  # TODO: still a case ?
        thresh: float = 3.5,
        min_periods_r: int = 1,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        The algorithm implements a 3-step outlier detection procedure for simultaneously
        flagging of higher dimensional data (dimensions > 3).

        In [1], the procedure is introduced and exemplified with an application on hydrological
        data. See the notes section for an overview over the algorithms basic steps.

        Parameters
        ----------
        trafo : default identity
            Transformation to be applied onto every column before scoring. For more fine-grained
            control, the data could also be transformed before :py:meth:`~saqc.SaQC.flagMVScores`
            is called.

        alpha :
            Level of significance by which it is tested, if an observations score might
            be drawn from another distribution than the majority of the data.

        n :
            Number of neighbors included in the scoring process for every datapoint.

        func : default sum
            Function that aggregates a value's k-smallest distances, returning a scalar score.

        iter_start :
            Value in ``[0,1]`` that determines which percentage of data is considered
            "normal". 0.5 results in the threshing algorithm to search only the upper 50%
            of the scores for the cut off point. (See reference section for more
            information)

        window :
            Only effective if :py:attr:`threshing` is set to ``'stray'``. Determines the
            size of the data partitions, the data is decomposed into. Each partition is checked
            seperately for outliers.
            Either given as an Offset String, denoting the windows temporal extension or
            as an integer, denoting the windows number of periods. ``NaN`` also count as periods.
            If ``None``, all data points share the same scoring window, which than equals the whole
            data.

        min_periods :
            Only effective if :py:attr:`threshing` is set to ``'stray'`` and :py:attr:`partition` is an integer.
            Minimum number of periods per :py:attr:`partition` that have to be present for a valid outlier
            detection to be made in this partition.

        stray_range :
            If not ``None``, it is tried to reduce the stray result onto single outlier components
            of the input :py:attr:`field`. The offset string denotes the range of the
            temporal surrounding to include into the MAD testing while trying to reduce
            flags.

        drop_flagged :
            Only effective when :py:attr:`stray_range` is not ``None``. Whether or not to drop flagged
            values from the temporal surroundings.

        thresh :
            Only effective when :py:attr:`stray_range` is not ``None``. The 'critical' value,
            controlling wheather the MAD score is considered referring to an outlier or
            not. Higher values result in less rigid flagging. The default value is widely
            considered apropriate in the literature.

        min_periods_r :
            Only effective when :py:attr:`stray_range` is not ``None``. Minimum number of measurements
            necessary in an interval to actually perform the reduction step.

        Notes
        -----
        The basic steps are:

        1. transforming

        The different data columns are transformed via timeseries transformations to
        (a) make them comparable and
        (b) make outliers more stand out.

        This step is usually subject to a phase of research/try and error. See [1] for more
        details.

        Note, that the data transformation as an built-in step of the algorithm, will likely
        get deprecated in the future. Its better to transform the data in a processing
        step, preceeding the multivariate flagging process. Also, by doing so, one gets
        mutch more control and variety in the transformation applied, since the `trafo`
        parameter only allows for application of the same transformation to all of the
        variables involved.

        2. scoring

        Every observation gets assigned a score depending on its k nearest neighbors. See
        the `scoring_method` parameter description for details on the different scoring
        methods. Furthermore [1] may give some insight in the pro and cons of the
        different methods.

        3. threshing

        The gaps between the (greatest) scores are tested for beeing drawn from the same
        distribution as the majority of the scores. If a gap is encountered, that,
        with sufficient significance, can be said to not be drawn from the same
        distribution as the one all the smaller gaps are drawn from, than the observation
        belonging to this gap, and all the observations belonging to gaps larger then
        this gap, get flagged outliers. See description of the `threshing` parameter for
        more details. Although [1] gives a fully detailed overview over the `stray`
        algorithm.

        References
        ----------
        [1]  Priyanga Dilini Talagala, Rob J. Hyndman & Kate Smith-Miles (2021):
             Anomaly Detection in High-Dimensional Data,
             Journal of Computational and Graphical Statistics, 30:2, 360-374,
             DOI: 10.1080/10618600.2020.1807997

        """

        # parameter deprecations

        if "partition" in kwargs:
            warnings.warn(
                """
                The parameter `partition` is deprecated and will be removed in version 3.0 of saqc.
                Please us the parameter `window` instead.'
                """,
                DeprecationWarning,
            )
            window = kwargs["partition"]

        if "partition_min" in kwargs:
            warnings.warn(
                """
                The parameter `partition_min` is deprecated and will be removed in version 3.0 of saqc.
                Please us the parameter `min_periods` instead.'
                """,
                DeprecationWarning,
            )
            min_periods = kwargs["partition_min"]

        if min_periods != 11:
            warnings.warn(
                """
                You were setting a customary value for the `min_periods` parameter: note that this parameter 
                does no longer refer to the reduction interval length, but now controls the number of periods 
                having to be present in an interval of size `window` (deprecated:`partition`) for the algorithm to be 
                performed in that interval.
                To alter the size of the reduction window, use the parameter `min_periods_r`. Changes readily apply. 
                Warning will be removed in saqc version 3.0.
                """,
                DeprecationWarning,
            )

        fields = toSequence(field)

        fields_ = []
        for f in fields:
            field_ = str(uuid.uuid4())
            self = self.copyField(field=f, target=field_)
            self = self.transform(field=field_, func=trafo, freq=window)
            fields_.append(field_)

        knn_field = str(uuid.uuid4())
        self = self.assignKNNScore(
            field=fields_,
            target=knn_field,
            n=n,
            func=func,
            freq=window,
            algorithm="ball_tree",
            min_periods=min_periods,
            **kwargs,
        )
        for field_ in fields_:
            self = self.dropField(field_)

        self = self.flagByStray(
            field=knn_field,
            freq=window,
            min_periods=min_periods,
            iter_start=iter_start,
            alpha=alpha,
            flag=flag,
            **kwargs,
        )

        self._data, self._flags = _evalStrayLabels(
            data=self._data,
            field=knn_field,
            target=fields,
            flags=self._flags,
            reduction_range=stray_range,
            reduction_drop_flagged=drop_flagged,
            reduction_thresh=thresh,
            reduction_min_periods=min_periods_r,
            flag=flag,
            **kwargs,
        )
        return self.dropField(knn_field)

    @flagging()
    def flagRaise(
        self: "SaQC",
        field: str,
        thresh: float,
        raise_window: str,
        freq: str,
        average_window: str | None = None,
        raise_factor: float = 2.0,
        slope: float | None = None,
        weight: float = 0.8,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        The function flags raises and drops in value courses, that exceed a certain threshold
        within a certain timespan.

        The parameter variety of the function is owned to the intriguing case of values, that
        "return" from outlierish or anomalious value levels and thus exceed the threshold,
        while actually being usual values.

        Notes
        -----
        The dataset is NOT supposed to be harmonized to a time series with an equidistant
        requency grid.

        The value :math:`x_{k}` of a time series :math:`x` with associated
        timestamps :math:`t_i`, is flagged a raise, if:

        1. There is any value :math:`x_{s}`, preceeding :math:`x_{k}` within
           :py:attr:`raise_window` range, so that
           :math:`M = |x_k - x_s | >`  :py:attr:`thresh` :math:`> 0`

        2. The weighted average :math:`\\mu^{*}` of the values, preceding :math:`x_{k}`
           within :py:attr:`average_window` range indicates, that :math:`x_{k}` does not
           return from an "outlierish" value course, meaning that
           :math:`x_k > \\mu^* + ( M` / :py:attr:`raise_factor` :math:`)`

        3. Additionally, if :py:attr:`slope` is not ``None``, :math:`x_{k}` is checked for being
           sufficiently divergent from its very predecessor :math:`x_{k-1}`, meaning that, it
           is additionally checked if:
           * :math:`x_k - x_{k-1} >` :py:attr:`slope`
           * :math:`t_k - t_{k-1} >` :py:attr:`weight` :math:`\\times` :py:attr:`freq`

        Parameters
        ----------
        thresh :
            The threshold, for the total rise (:py:attr:`thresh` ``> 0``), or total drop
            (:py:attr:`thresh` ``< 0``), value courses must not exceed within a timespan
            of length :py:attr:`raise_window`.

        raise_window :
            An offset string, determining the timespan, the rise/drop thresholding refers
            to. Window is inclusively defined.

        freq :
            An offset string, determining the frequency, the timeseries to flag is supposed
            to be sampled at. The window is inclusively defined.

        average_window :
            See condition (2) of the description given in the Notes. Window is
            inclusively defined, defaults to 1.5 times the size of :py:attr:`raise_window`.

        raise_factor :
            See condition (2).

        slope :
            See condition (3).

        weight :
            See condition (3).
        """
        # prepare input args
        dataseries = self._data[field].dropna()
        raise_window_td = pd.Timedelta(raise_window)
        freq_dt = pd.Timedelta(freq)
        if slope is not None:
            slope = np.abs(slope)

        if average_window is None:
            average_window = 1.5 * raise_window_td

        if thresh < 0:
            dataseries *= -1
            thresh *= -1

        def raise_check(x, thresh):
            test_set = x[-1] - x[0:-1]
            max_val = np.max(test_set)
            if max_val >= thresh:
                return max_val
            else:
                return np.nan

        def custom_rolling_mean(x):
            return np.sum(x[:-1])

        # get invalid-raise/drop mask:
        raise_series = dataseries.rolling(raise_window_td, min_periods=2, closed="both")

        numba_boost = True
        if numba_boost:
            raise_check_boosted = numba.jit(raise_check, nopython=True)
            raise_series = raise_series.apply(
                raise_check_boosted, args=(thresh,), raw=True, engine="numba"
            )
        else:
            raise_series = raise_series.apply(raise_check, args=(thresh,), raw=True)

        if raise_series.isna().all():
            return self

        # "unflag" values of insufficient deviation to their predecessors
        if slope is not None:
            w_mask = (
                pd.Series(dataseries.index).diff().dt.total_seconds()
                / freq_dt.total_seconds()
            ) > weight
            slope_mask = np.abs(dataseries.diff()) < slope
            to_unflag = raise_series.notna() & w_mask.values & slope_mask
            raise_series[to_unflag] = np.nan

        # calculate and apply the weighted mean weights (pseudo-harmonization):
        weights = (
            pd.Series(dataseries.index).diff(periods=2).shift(-1).dt.total_seconds()
            / freq_dt.total_seconds()
            / 2
        )

        weights.iloc[0] = 0.5 + (
            pd.Timestamp(dataseries.index[1]) - pd.Timestamp(dataseries.index[0])
        ).total_seconds() / (freq_dt.total_seconds() * 2)

        weights.iloc[-1] = 0.5 + (
            pd.Timestamp(dataseries.index[-1]) - pd.Timestamp(dataseries.index[-2])
        ).total_seconds() / (freq_dt.total_seconds() * 2)

        weights[weights > 1.5] = 1.5
        weights.index = dataseries.index
        weighted_data = dataseries.mul(weights)

        # rolling weighted mean calculation
        weighted_rolling_mean = weighted_data.rolling(
            average_window, min_periods=2, closed="both"
        )
        weights_rolling_sum = weights.rolling(
            average_window, min_periods=2, closed="both"
        )
        if numba_boost:
            custom_rolling_mean_boosted = numba.jit(custom_rolling_mean, nopython=True)
            weighted_rolling_mean = weighted_rolling_mean.apply(
                custom_rolling_mean_boosted, raw=True, engine="numba"
            )
            weights_rolling_sum = weights_rolling_sum.apply(
                custom_rolling_mean_boosted, raw=True, engine="numba"
            )
        else:
            weighted_rolling_mean = weighted_rolling_mean.apply(
                custom_rolling_mean, raw=True
            )
            weights_rolling_sum = weights_rolling_sum.apply(
                custom_rolling_mean, raw=True, engine="numba"
            )

        weighted_rolling_mean = weighted_rolling_mean / weights_rolling_sum
        # check means against critical raise value:
        to_flag = dataseries >= weighted_rolling_mean + (raise_series / raise_factor)
        to_flag &= raise_series.notna()
        self._flags[to_flag[to_flag].index, field] = flag

        return self

    @flagging()
    def flagMAD(
        self: "SaQC",
        field: str,
        window: str | int | None = None,
        z: float = 3.5,
        min_residuals: int | None = None,
        min_periods: int | None = None,
        center: bool = False,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag outiers using the modified Z-score outlier detection method.

        See references [1] for more details on the algorithm.

        Note
        ----
        Data needs to be sampled at a regular equidistant time grid.

        Parameters
        ----------
        window :
            Size of the window. Either given as an Offset String, denoting the window's temporal extension or
            as an integer, denoting the window's number of periods. ``NaN`` also count as periods.
            If ``None``, all data points share the same scoring window, which than equals the whole data.
        z :
            The value the Z-score is tested against. Defaulting to ``3.5`` (Recommendation of [1])
        min_periods :
            Minimum number of valid meassurements in a scoring window, to consider the resulting score valid.
        center :
            Weather or not to center the target value in the scoring window. If ``False``, the
            target value is the last value in the window.

        References
        ----------
        [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
        """

        self = self.flagZScore(
            field,
            window=window,
            thresh=z,
            min_residuals=min_residuals,
            model_func=np.median,
            norm_func=lambda x: median_abs_deviation(
                x, scale="normal", nan_policy="omit"
            ),
            center=center,
            min_periods=min_periods,
            flag=flag,
        )

        return self

    @flagging()
    def flagOffset(
        self: "SaQC",
        field: str,
        tolerance: float,
        window: int | str,
        thresh: float | None = None,
        thresh_relative: float | None = None,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        A basic outlier test that works on regularly and irregularly sampled data.

        The test classifies values/value courses as outliers by detecting not only a rise
        in value, but also, by checking for a return to the initial value level.

        Notes
        -----
        This definition of a "spike" not only includes one-value outliers, but also plateau-ish value courses.

        Values :math:`x_n, x_{n+1}, .... , x_{n+k}` of a timeseries :math:`x` with
        associated timestamps :math:`t_n, t_{n+1}, .... , t_{n+k}` are considered spikes, if:
        1. :math:`|x_{n-1} - x_{n + s}| >` :py:attr:`thresh`, for all :math:`s \\in [0,1,2,...,k]`
        2. if :py:attr:`thresh_relative` > 0, :math:`x_{n + s} > x_{n - 1}*(1+` :py:attr:`thresh_relative` :math:`)`
        3. if :py:attr:`thresh_relative` < 0, :math:`x_{n + s} < x_{n - 1}*(1+` :py:attr:`thresh_relative` :math:`)`
        4. :math:`|x_{n-1} - x_{n+k+1}| <` :py:attr:`tolerance`
        5. :math:`|t_{n-1} - t_{n+k+1}| <` :py:attr:`window`


        Parameters
        ----------
        tolerance :
            Maximum difference allowed between the value, directly preceding and the value directly
            succeeding an offset to trigger flagging of the offsetting values. See condition (4).
        window :
            Maximum length allowed for offset value courses, to trigger flagging of the offsetting values.
            See condition (5). Integer defined window length are only allowed for regularly sampled
            timeseries.
        thresh :
            Minimum difference between a value and its successors, to consider the successors an anomalous
            offset group. See condition (1). If ``None``, condition (1) is not tested.
        thresh_relative :
            Minimum relative change between a value and its successors, to consider the successors an anomalous
            offset group. See condition (2). If ``None``, condition (2) is not tested.

        Examples
        --------
        Below picture gives an abstract interpretation of the parameter interplay in case of a positive
        value jump, initialising an offset course.

        .. figure:: /resources/images/flagOffsetPic.png

           The four values marked red, are flagged, because (1) the initial value jump *exceeds* the value
           given by :py:attr:`thresh`, (2) the temporal extension of the group does *not exceed* the range
           given by `window` and (3) the returning value after the group, lies *within* the value range
           determined by :py:attr:`tolerance`


        .. plot::
           :context:
           :include-source: False

           import matplotlib
           import saqc
           import pandas as pd
           data = pd.DataFrame({'data':np.array([5,5,8,16,17,7,4,4,4,1,1,4])}, index=pd.date_range('2000',freq='1H', periods=12))


        Lets generate a simple, regularly sampled timeseries with an hourly sampling rate and generate an
        :py:class:`saqc.SaQC` instance from it.

        .. doctest:: flagOffsetExample

           >>> import saqc
           >>> data = pd.DataFrame({'data':np.array([5,5,8,16,17,7,4,4,4,1,1,4])}, index=pd.date_range('2000',freq='1H', periods=12))
           >>> data
                                data
           2000-01-01 00:00:00     5
           2000-01-01 01:00:00     5
           2000-01-01 02:00:00     8
           2000-01-01 03:00:00    16
           2000-01-01 04:00:00    17
           2000-01-01 05:00:00     7
           2000-01-01 06:00:00     4
           2000-01-01 07:00:00     4
           2000-01-01 08:00:00     4
           2000-01-01 09:00:00     1
           2000-01-01 10:00:00     1
           2000-01-01 11:00:00     4
           >>> qc = saqc.SaQC(data)

        Now we are applying :py:meth:`~saqc.SaQC.flagOffset` and try to flag offset courses, that dont extend
        longer than *6 hours* in time (:py:attr:`window`) and that have an initial value jump higher than ``2``
        (:py:attr:`thresh`), and that do return to the initial value level within a tolerance of ``1.5``
        (:py:attr:`tolerance`).

        .. doctest:: flagOffsetExample

           >>> qc = qc.flagOffset("data", thresh=2, tolerance=1.5, window='6H')
           >>> qc.plot('data')  # doctest: +SKIP

        .. plot::
           :context: close-figs
           :include-source: False

           >>> qc = saqc.SaQC(data)
           >>> qc = qc.flagOffset("data", thresh=2, tolerance=1.5, window='6H')
           >>> qc.plot('data')  # doctest: +SKIP

        Note, that both, negative and positive jumps are considered starting points of negative or positive
        offsets. If you want to impose the additional condition, that the initial jump must exceed
        +90%* of the value level, you can additionally set the :py:attr:`thresh_relative` parameter:

        .. doctest:: flagOffsetExample

           >>> qc = qc.flagOffset("data", thresh=2, thresh_relative=.9, tolerance=1.5, window='6H')
           >>> qc.plot('data') # doctest:+SKIP

        .. plot::
           :context: close-figs
           :include-source: False

           >>> qc = saqc.SaQC(data)
           >>> qc = qc.flagOffset("data", thresh=2, thresh_relative=.9, tolerance=1.5, window='6H')
           >>> qc.plot('data')  # doctest: +SKIP

        Now, only positive jumps, that exceed a value gain of +90%* are considered starting points of offsets.

        In the same way, you can aim for only negative offsets, by setting a negative relative threshold.
        The below example only flags offsets, that fall off by at least *50%* in value, with an absolute
        value drop of at least 2.

        .. doctest:: flagOffsetExample

           >>> qc = qc.flagOffset("data", thresh=2, thresh_relative=-.5, tolerance=1.5, window='6H')
           >>> qc.plot('data') # doctest:+SKIP

        .. plot::
           :context: close-figs
           :include-source: False

           >>> qc = saqc.SaQC(data)
           >>> qc = qc.flagOffset("data", thresh=2, thresh_relative=-.5, tolerance=1.5, window='6H')
           >>> qc.plot('data')  # doctest: +SKIP
        """
        if (thresh is None) and (thresh_relative is None):
            raise ValueError(
                "At least one of parameters 'thresh' and 'thresh_relative' has to be given. Got 'thresh'=None, "
                "'thresh_relative'=None instead."
            )
        if thresh is None:
            thresh = 0

        dat = self._data[field].dropna()
        if thresh_relative is not None:
            rel_jumps = np.sign(thresh_relative) * dat > np.sign(
                thresh_relative
            ) * dat.shift(+1) * (1 + thresh_relative)

        data_diff = dat.diff()
        initial_jumps = data_diff.abs() > thresh
        if thresh_relative:
            initial_jumps &= rel_jumps
        return_in_time = (
            dat[::-1]
            .rolling(window, min_periods=2)
            .apply(lambda x: np.abs(x[-1] - x[:-1]).min() < tolerance, raw=True)[::-1]
            .astype(bool)
        )
        return_in_time = return_in_time & initial_jumps.reindex(
            dat.index, fill_value=False
        ).shift(-1, fill_value=False)
        offset_start_candidates = dat[return_in_time]
        win_delta = pd.Timedelta(window)
        corners = pd.Series(False, index=dat.index)
        to_flag = pd.Series(False, index=dat.index)
        ns = pd.Timedelta("1ns")
        for c in zip(offset_start_candidates.index, offset_start_candidates.values):
            ret = (dat[c[0]] - dat[c[0] + ns : c[0] + win_delta]).abs()[1:] < tolerance
            if not ret.empty:
                r = ret.idxmax()
                chunk = dat[c[0] : r]
                sgn = np.sign(chunk[1] - c[1])
                t_val = ((chunk[1:-1] - c[1]) * sgn > thresh).all()
                r_val = True
                if thresh_relative:
                    r_val = (
                        np.sign(thresh_relative) * chunk[1:-1]
                        > np.sign(thresh_relative) * c[1] * (1 + thresh_relative)
                    ).all()
                if t_val and r_val and (not corners[c[0]]):
                    flag_i = dat[c[0] + ns : chunk.index[-1] - ns].index
                    to_flag[flag_i] = True
                    corners.loc[flag_i[-1]] = True
        to_flag = to_flag.reindex(self._data[field].index, fill_value=False)

        self._flags[to_flag, field] = flag
        return self

    @flagging()
    def flagByGrubbs(
        self: "SaQC",
        field: str,
        window: str | int,
        alpha: float = 0.05,
        min_periods: int = 8,
        pedantic: bool = False,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag outliers using the Grubbs algorithm.

        See [1] for more information on the grubbs tests definition.

        The (two-sided) test gets applied to data chunks of size :py:attr:`window`. The
        tests will be iterated chunkwise until no more outliers are detected.

        Note
        ----
        * The data is expected to be normally distributed!
        * The test performs poorly for small data chunks, resulting in considerable
          overflagging. Select :py:attr:`window` such that every data chunk contains at
          least 8 values and also adjust the :py:attr:`min_periods` values accordingly.

        Parameters
        ----------

        window :
            Size of the testing window.
            If an integer, the fixed number of observations used for each window.
            If an offset string the time period of each window.

        alpha :
            Level of significance, the grubbs test is to be performed at. Must be between 0 and 1.

        min_periods :
            Minimum number of values needed in a :py:attr:`window` in order to perform the grubs test.
            Ignored if :py:attr:`window` is an integer.

        pedantic :
            If ``True``, every value gets checked twice. First in the initial rolling :py:attr:`window`
            and second in a rolling window that is lagging by :py:attr:`window` / 2. Recommended to avoid
            false positives at the window edges. Ignored if :py:attr:`window` is an offset string.

        References
        ----------
        introduction to the grubbs test:

        [1] https://en.wikipedia.org/wiki/Grubbs%27s_test_for_outliers
        """
        datcol = self._data[field].copy()
        rate = getFreqDelta(datcol.index)

        # if timeseries that is analyzed, is regular,
        # window size can be transformed to a number of periods:
        if rate and isinstance(window, str):
            window = pd.Timedelta(window) // rate

        to_group = pd.DataFrame(data={"ts": datcol.index, "data": datcol})
        to_flag = pd.Series(False, index=datcol.index)

        # period number defined test intervals
        if isinstance(window, int):
            grouper_series = pd.Series(
                data=np.arange(0, len(datcol)), index=datcol.index
            )
            grouper_series_lagged = grouper_series + (window / 2)
            grouper_series = grouper_series.transform(lambda x: x // window)
            grouper_series_lagged = grouper_series_lagged.transform(
                lambda x: x // window
            )
            partitions = to_group.groupby(grouper_series)
            partitions_lagged = to_group.groupby(grouper_series_lagged)

        # offset defined test intervals:
        else:
            partitions = to_group.groupby(pd.Grouper(freq=window))
            partitions_lagged = []

        for _, partition in partitions:
            if partition.shape[0] > min_periods:
                detected = smirnov_grubbs.two_sided_test_indices(
                    partition["data"].values, alpha=alpha
                )
                detected = partition["ts"].iloc[detected]
                to_flag[detected.index] = True

        if isinstance(window, int) and pedantic:
            to_flag_lagged = pd.Series(False, index=datcol.index)

            for _, partition in partitions_lagged:
                if partition.shape[0] > min_periods:
                    detected = smirnov_grubbs.two_sided_test_indices(
                        partition["data"].values, alpha=alpha
                    )
                    detected = partition["ts"].iloc[detected]
                    to_flag_lagged[detected.index] = True

            to_flag &= to_flag_lagged

        self._flags[to_flag, field] = flag
        return self

    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],
        multivariate=True,
        handles_target=False,
        docstring={"field": DOC_TEMPLATES["field"]},
    )
    def flagCrossStatistics(
        self: "SaQC",
        field: Sequence[str],
        thresh: float,
        method: Literal["modZscore", "Zscore"] = "modZscore",
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Function checks for outliers relatively to the "horizontal" input data axis.

        Notes
        -----
        The input variables dont necessarily have to be aligned. If the variables are unaligned, scoring
        and flagging will only be performed on the subset of indices shared among all input variables.

        For :py:attr:`field` :math:`=[f_1,f_2,...,f_N]` and timestamps :math:`[t_1,t_2,...,t_K]`,
        the following steps are taken for outlier detection:

        1. All timestamps :math:`t_i`, where there is one :math:`f_k`, with :math:`data[f_K]` having no
           entry at :math:`t_i`, are excluded from the following process (inner join of the :math:`f_i` fields.)
        2. for every :math:`0 <= i <= K`, the value
           :math:`m_j = median(\\{data[f_1][t_i], data[f_2][t_i], ..., data[f_N][t_i]\\})` is calculated
        3. for every :math:`0 <= i <= K`, the set
           :math:`\\{data[f_1][t_i] - m_j, data[f_2][t_i] - m_j, ..., data[f_N][t_i] - m_j\\}` is tested for
           outliers with the specified algorithm (:py:attr:`method` parameter).

        Parameters
        ----------
        thresh :
            Threshold which the outlier score of an value must exceed, for being flagged an outlier.

        method :
            Method used for calculating the outlier scores.

            * ``'modZscore'``: Median based "sigma"-ish approach. See References [1].
            * ``'Zscore'``: Score values by how many times the standard deviation they differ from the
              median. See References [1].


        References
        ----------
        [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
        """
        warnings.warn(
            "The method `flagCrossStatistics` will be deprecated in a future version of saqc",
            PendingDeprecationWarning,
        )

        fields = toSequence(field)

        df = self._data[fields].to_pandas(how="inner")

        if isinstance(method, str):
            if method == "modZscore":
                MAD_series = df.subtract(df.median(axis=1), axis=0).abs().median(axis=1)
                diff_scores = (
                    (0.6745 * (df.subtract(df.median(axis=1), axis=0)))
                    .divide(MAD_series, axis=0)
                    .abs()
                )

            elif method == "Zscore":
                diff_scores = (
                    df.subtract(df.mean(axis=1), axis=0)
                    .divide(df.std(axis=1), axis=0)
                    .abs()
                )

            else:
                raise ValueError(method)

        else:
            try:
                stat = getattr(df, method.__name__)(axis=1)
            except AttributeError:
                stat = df.aggregate(method, axis=1)

            diff_scores = df.subtract(stat, axis=0).abs()

        mask = diff_scores > thresh
        if not mask.empty:
            for f in fields:
                m = mask[f].reindex(index=self._flags[f].index, fill_value=False)
                self._flags[m, f] = flag

        return self

    @flagging()
    def flagZScore(
        self: "SaQC",
        field: str,
        window: str | int | None = None,
        thresh: float = 3,
        min_residuals: int | None = None,
        min_periods: int | None = None,
        model_func: Callable[[np.ndarray | pd.Series], float] = np.nanmean,
        norm_func: Callable[[np.ndarray | pd.Series], float] = np.nanstd,
        center: bool = True,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Flag data where its (rolling) Zscore exceeds a threshold.

        The function implements flagging derived from a basic Zscore calculation. To handle non
        stationary data, the Zscoring can be applied with a rolling window. Therefor, the function
        allows for a minimum residual to be specified in order to mitigate overflagging in local
        regimes of low variance.

        See the Notes section for a detailed overview of the calculation

        Parameters
        ----------
        window :
            Size of the window. Either determined via an Offset String, denoting the windows temporal
            extension or by an integer, denoting the windows number of periods. ``NaN`` also count as
            periods. If ``None`` is passed, all data points share the same scoring window, which than
            equals the whole data.
        thresh :
            Cutoff level for the Zscores, above which associated points are marked as outliers.
        min_residuals :
            Minimum residual value points must have to be considered outliers.
        min_periods :
            Minimum number of valid meassurements in a scoring window, to consider the resulting score valid.
        model_func : default mean
            Function to calculate the center moment in every window.
        norm_func : default std
            Function to calculate the scaling for every window.
        center :
            Weather or not to center the target value in the scoring window. If ``False``, the
            target value is the last value in the window.

        Notes
        -----
        Steps of calculation:

        1. Consider a window :math:`W` of successive points :math:`W = x_{1},...x_{w}`
           containing the value :math:`y_{K}` which is to be checked.
           (The index of :math:`K` depends on the selection of the parameter :py:attr:`center`.)
        2. The "moment" :math:`M` for the window gets calculated via :math:`M=` :py:attr:`model_func` :math:`(W)`.
        3. The "scaling" :math:`N` for the window gets calculated via :math:`N=` :py:attr:`norm_func` :math:`(W)`.
        4. The "score" :math:`S` for the point :math:`x_{k}` gets calculated via :math:`S=(x_{k} - M) / N`.
        5. Finally, :math:`x_{k}` gets flagged, if :math:`|S| >` :py:attr:`thresh` and
           :math:`|M - x_{k}| >=` :py:attr:`min_residuals`.
        """
        datser = self._data[field]
        if min_residuals is None:
            min_residuals = 0

        score, model, _ = _univarScoring(
            datser,
            window=window,
            norm_func=norm_func,
            model_func=model_func,
            center=center,
            min_periods=min_periods,
        )
        to_flag = (score.abs() > thresh) & ((model - datser).abs() >= min_residuals)
        self._flags[to_flag, field] = flag
        return self


def _evalStrayLabels(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    target: Sequence[str],
    reduction_range: Optional[str] = None,
    reduction_drop_flagged: bool = False,  # TODO: still a case ?
    reduction_thresh: float = 3.5,
    reduction_min_periods: int = 1,
    at_least_one: bool = True,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function "reduces" an observations flag to components of it, by applying MAD
    (See references) test onto every components temporal surrounding.

    Parameters
    ----------
    data :
        A dictionary of pandas.Series, holding all the data.

    field :
        The fieldname of the column, holding the labels to be evaluated.

    flags :
        Container to store quality flags to data.

    target :
        A list of strings, holding the column names of the variables, the stray labels
        shall be projected onto.

    val_frame :
        Input NxM DataFrame of observations, where N is the number of observations and
        M the number of components per observation.

    to_flag_frame :
        Input dataframe of observations to be tested, where N is the number of
        observations and M the number of components per observation.

    reduction_range :
        An offset string, denoting the range of the temporal surrounding to include
        into the MAD testing. If ``None`` is passed, no testing will be performed and
        all targets will have the stray flag projected.

    reduction_drop_flagged :
        Wheather or not to drop flagged values other than the value under test, from the
        temporal surrounding before checking the value with MAD.

    reduction_thresh :
        The `critical` value, controlling wheather the MAD score is considered
        referring to an outlier or not. Higher values result in less rigid flagging.
        The default value is widely used in the literature. See references section
        for more details ([1]).

    at_least_one :
        If none of the variables, the outlier label shall be reduced to, is an outlier
        with regard to the test, all (True) or none (False) of the variables are flagged

    flag : float, default BAD
        flag to set.

    References
    ----------
    [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
    """
    val_frame = data[target].to_pandas()
    stray_detects = flags[field] > UNFLAGGED
    stray_detects = stray_detects[stray_detects]
    to_flag_frame = pd.DataFrame(False, columns=target, index=stray_detects.index)

    if reduction_range is None:
        for field in to_flag_frame.columns:
            flags[to_flag_frame.index, field] = flag
        return data, flags

    for var in target:
        for index in enumerate(to_flag_frame.index):
            index_slice = slice(
                index[1] - pd.Timedelta(reduction_range),
                index[1] + pd.Timedelta(reduction_range),
            )
            test_slice = val_frame[var][index_slice].dropna()

            # check, wheather value under test is sufficiently centered:
            first = test_slice.first_valid_index()
            last = test_slice.last_valid_index()
            min_range = pd.Timedelta(reduction_range) / 4

            if (
                pd.Timedelta(index[1] - first) < min_range
                or pd.Timedelta(last - index[1]) < min_range
            ):
                polydeg = 0
            else:
                polydeg = 2

            if reduction_drop_flagged:
                test_slice = test_slice.drop(to_flag_frame.index, errors="ignore")

            if len(test_slice) < reduction_min_periods:
                to_flag_frame.loc[index[1], var] = True
                continue

            x = test_slice.index.values.astype(float)
            x_0 = x[0]
            x = (x - x_0) / 10**12

            polyfitted = poly.polyfit(y=test_slice.values, x=x, deg=polydeg)

            testval = poly.polyval(
                (float(index[1].to_numpy()) - x_0) / 10**12, polyfitted
            )
            testval = val_frame[var][index[1]] - testval

            resids = test_slice.values - poly.polyval(x, polyfitted)
            med_resids = np.median(resids)
            MAD = np.median(np.abs(resids - med_resids))
            crit_val = 0.6745 * (abs(med_resids - testval)) / MAD

            if crit_val > reduction_thresh:
                to_flag_frame.loc[index[1], var] = True

    if at_least_one:
        to_flag_frame[~to_flag_frame.any(axis=1)] = True

    for field in to_flag_frame.columns:
        col = to_flag_frame[field]
        flags[col[col].index, field] = flag

    return data, flags



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict

import numpy as np
import pandas as pd

from saqc import BAD, FILTER_ALL, GOOD, UNFLAGGED
from saqc.core import DictOfSeries, Flags
from saqc.lib.types import ExternalFlag

ForwardMap = Dict[ExternalFlag, float]
BackwardMap = Dict[float, ExternalFlag]


class TranslationScheme:  # pragma: no cover
    @property
    @abstractmethod
    def DFILTER_DEFAULT(self):
        pass

    @abstractmethod
    def __call__(self, flag: ExternalFlag) -> float:
        pass

    @abstractmethod
    def toInternal(self, flags: pd.DataFrame | DictOfSeries) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """
        pass

    @abstractmethod
    def toExternal(self, flags: Flags, attrs: dict | None = None) -> DictOfSeries:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        attrs : dict or None, default None
            global meta information of saqc-object

        Returns
        -------
        pd.DataFrame
        """
        pass


class MappingScheme(TranslationScheme):
    """
    This class provides the basic translation mechanism and should serve as
    a base class for most other translation scheme.

    The general translation is realized through dictionary lookups, although
    we might need to extend this logic to also allow calls to translation
    functions in the future. Currently, at least one `dict` defining the
    'forward' translation  from 'user flags' -> 'internal flags' needs to be
    provided.
    Optionally a second `dict` can be passed to map 'internal flags' -> 'user flags',
    if the latter is not given, this 'backward' translation is inferred as
    the inverse of the 'forward' translation.

    The translation mechanism imposes a few restrictions:

    - The scheme must be well definied, i.e. we need a backward translation for
      every forward translation (each value in `self._forward` needs a key in
      `self._backward`).
    - We need translations for the special flags:
      * `saqc.constants.UNFLAGGED`
      * `saqc.constants.BAD`

    . That implies, that every valid translation scheme
      provides at least one user flag that maps to `BAD` and one that maps to
      `UNFLAGGED`.
    """

    # (internal) threshold flag above which values will be masked
    DFILTER_DEFAULT: float = FILTER_ALL

    # additional arguments and default values the translation scheme accepts
    ARGUMENTS: Dict[str, Any] = {}

    def __init__(self, forward: ForwardMap, backward: BackwardMap):
        """
        Parameters
        ----------
        forward : dict
            A mapping defining the forward translation of scalar flag values

        backward : dict
            A mapping defining the backward translation of scalar flag values

        Note
        ----
        `backward` needs to provide a mapping for the two special flags
        `saqc.constants.UNFLAGGED`, `saqc.constants.BAD`
        """
        if UNFLAGGED not in backward or BAD not in backward:
            raise ValueError(
                f"need translations for the special flags `UNFLAGGED` ({UNFLAGGED})"
                f" and `BAD` ({BAD})"
            )
        self._forward = forward
        self._backward = backward

    @staticmethod
    def _translate(
        flags: Flags | pd.DataFrame | pd.Series | DictOfSeries,
        trans_map: ForwardMap | BackwardMap,
    ) -> DictOfSeries:
        """
        Translate a given flag data structure to another according to the
        mapping given in `trans_map`

        Parameters
        ----------
        flags : Flags, pd.DataFrame
            The flags to translate

        Returns
        -------
        DictOfSeries
        """
        if isinstance(flags, pd.Series):
            flags = flags.to_frame()

        out = DictOfSeries()
        expected = pd.Index(trans_map.values())
        for field in flags.columns:
            out[field] = flags[field].replace(trans_map)
            diff = pd.Index(out[field]).difference(expected)
            if not diff.empty:
                raise ValueError(
                    f"following flag values could not be "
                    f"translated: {diff.drop_duplicates().to_list()}"
                )
        return out

    def __call__(self, flag: ExternalFlag) -> float:
        """
        Translate a scalar 'external flag' to an 'internal flag'

        Parameters
        ----------
        flag : float, int, str
            The external flag to translate

        Returns
        -------
        float
        """
        if flag not in self._forward:
            if flag not in self._backward:
                raise ValueError(f"invalid flag: {flag}")
            return float(flag)
        return float(self._forward[flag])

    def toInternal(self, flags: pd.DataFrame | DictOfSeries | pd.Series) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """
        return Flags(self._translate(flags, self._forward))

    def toExternal(
        self,
        flags: Flags,
        attrs: dict | None = None,
    ) -> DictOfSeries:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        attrs : dict or None, default None
            global meta information of saqc-object

        Returns
        -------
        pd.DataFrame
        """
        out = self._translate(flags, self._backward)
        out.attrs = attrs or {}
        return out


class FloatScheme(TranslationScheme):

    """
    Acts as the default Translator, provides a changeable subset of the
    internal float flags
    """

    DFILTER_DEFAULT: float = FILTER_ALL

    def __call__(self, flag: float | int) -> float:
        try:
            return float(flag)
        except (TypeError, ValueError, OverflowError):
            raise ValueError(f"invalid flag, expected a numerical value, got: {flag}")

    def toInternal(self, flags: pd.DataFrame | DictOfSeries) -> Flags:
        try:
            return Flags(flags.astype(float))
        except (TypeError, ValueError, OverflowError):
            raise ValueError(
                f"invalid flag(s), expected a collection of numerical values, got: {flags}"
            )

    def toExternal(self, flags: Flags, attrs: dict | None = None) -> DictOfSeries:
        out = DictOfSeries(flags)
        out.attrs = attrs or {}
        return out



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import typing
import warnings
from typing import DefaultDict, Dict, Iterable, Mapping, Tuple, Type, Union, overload

import numpy as np
import pandas as pd

from saqc.core import DictOfSeries, History

_VAL = Union[pd.Series, History]
DictLike = Union[
    pd.DataFrame,
    DictOfSeries,
    Dict[str, _VAL],
    DefaultDict[str, _VAL],
]

_Field = str
SelectT = Union[
    _Field,
    Tuple[np.ndarray, _Field],
    Tuple[pd.Series, _Field],
    Tuple[pd.Index, _Field],
    Tuple[slice, _Field],
]
ValueT = Union[pd.Series, Iterable, float]


class _HistAccess:
    def __init__(self, obj: Flags):
        self._obj = obj

    def __getitem__(self, key: str) -> History:
        return self._obj._data[key]

    def __setitem__(self, key: str, value: History):
        if not isinstance(value, History):
            raise TypeError("Not a History")

        self._obj._validateHistForFlags(value)
        self._obj._data[key] = value


class Flags:
    """
    SaQC's flags container.

    This container class holds the quality flags associated with the data. It hold key-value pairs, where
    the key is the name of the column and the value is a ``pandas.Series`` of flags. The index of the series
    and the key-value pair can be assumed to be immutable, which means, only the *values* of the series can
    be change, once the series exist.
    In other words: **an existing column can not be overwritten by a column with a different index.**

    The flags can be accessed via ``__getitem__`` and ``__setitem__``, in real life known as the `[]`-operator.

    For the curious:
        Under the hood, the series are stored in a `history`, which allows the advanced user to retrieve all flags
        once was set in this object, but in the most cases this is irrelevant. For simplicity one can safely assume,
        that this class works just stores the flag-series one sets.

    See Also
    --------
    initFlagsLike : create a Flags instance, with same dimensions as a reference object.
    History : class that actually store the flags

    Examples
    --------

    We create an empty instance, by calling ``Flags`` without any arguments and then add a column to it.

    .. doctest:: exampleFlags

       >>> from saqc import UNFLAGGED, BAD, DOUBTFUL, Flags
       >>> flags = Flags()
       >>> flags
       Empty Flags

    .. doctest:: exampleFlags

       >>> flags['v0'] = pd.Series([BAD,BAD,UNFLAGGED], dtype=float)
       >>> flags # doctest:+NORMALIZE_WHITESPACE
             v0 |
       ======== |
       0  255.0 |
       1  255.0 |
       2   -inf |
       <BLANKLINE>

    Once the column exist, we cannot overwrite it anymore, with a different series.

    .. doctest:: exampleFlags

       >>> flags['v0'] = pd.Series([666.], dtype=float) # doctest:+ELLIPSIS
       Traceback (most recent call last):
         ...
       ValueError: Index does not match


    But if we pass a series, which index match it will work,
    because the series now is interpreted as value-to-set.

    .. doctest:: exampleFlags

       >>> flags['v0'] = pd.Series([DOUBTFUL,np.nan,DOUBTFUL], dtype=float)
       >>> flags # doctest:+NORMALIZE_WHITESPACE
             v0 |
       ======== |
       0   25.0 |
       1  255.0 |
       2   25.0 |
       <BLANKLINE>

    As we see above, the column now holds a combination from the values from the
    first and the second set. This is, because ``numpy.nan`` was used.
    We can inspect all the updates that was
    made by looking in the history.

    .. doctest:: exampleFlags

       >>> flags['v0'] = pd.Series([DOUBTFUL, np.nan, DOUBTFUL], dtype=float)
       >>> flags.history['v0'] # doctest:+NORMALIZE_WHITESPACE
             0     1     2
       0  255.0  25.0  25.0
       1  255.0   nan   nan
       2   -inf  25.0  25.0


    As we see now, the second call sets ``25.0`` and shadows (represented by the parentheses) ``(255.0)`` in the
    first row and ``(-inf)`` in the last, but in the second row ``255.0`` still is valid, because it was
    `not touched` by the set.

    It is also possible to set values by a mask, which can be interpreted as condidional setting.
    Imagine we want to `reset` all flags to ``0.`` if the existing flags are lower that ``255.``.

    .. doctest:: exampleFlags

       >>> mask = flags['v0'] < BAD
       >>> mask # doctest:+NORMALIZE_WHITESPACE
       0     True
       1    False
       2     True
       dtype: bool

    .. doctest:: exampleFlags

       >>> flags[mask, 'v0'] = 0
       >>> flags # doctest:+NORMALIZE_WHITESPACE
             v0 |
       ======== |
       0    0.0 |
       1  255.0 |
       2    0.0 |
       <BLANKLINE>

    The objects you can pass as a row selector (``flags[rows, column]``) are:

    - boolen arraylike, with or without index. Must have same length than the undeliing series.
    - slices working on the index
    - ``pd.Index``, which must be a subset of the existing index

    For example, to set `all` values to a scalar value, use a Null-slice:

    .. doctest:: exampleFlags

       >>> flags[:, 'v0'] = 99.0
       >>> flags # doctest:+NORMALIZE_WHITESPACE
            v0 |
       ======= |
       0  99.0 |
       1  99.0 |
       2  99.0 |
       <BLANKLINE>


    After all calls presented here, the history looks like this:

    .. doctest:: exampleFlags

       >>> flags.history['v0']
             0     1     2    3     4
       0  255.0  25.0  25.0  0.0  99.0
       1  255.0   nan   nan  nan  99.0
       2   -inf  25.0  25.0  0.0  99.0
    """

    def __init__(self, raw_data: DictLike | Flags | None = None, copy: bool = False):
        self._data: dict[str, History]

        if raw_data is None:
            raw_data = {}

        if isinstance(raw_data, Flags):
            if copy:
                raw_data = raw_data.copy()
            self._data = raw_data._data

        else:
            self._data = self._initFromRaw(raw_data, copy)

    @staticmethod
    def _initFromRaw(data: Mapping, copy: bool) -> Dict[str, History]:
        """
        init from dict-like: keys are flag column, values become
        initial columns of history(s).
        """
        result = {}

        for k, item in data.items():
            if not isinstance(k, str):
                raise ValueError("column names must be string")
            if k in result:
                raise ValueError("raw_data must not have duplicate keys")

            # a passed History is not altered. So if the passed History
            # does not fit for Flags, we fail hard.
            if isinstance(item, History):
                Flags._validateHistForFlags(item, colname=k)
                if copy:
                    item = item.copy()
                result[k] = item
                continue
            if not isinstance(item, pd.Series):
                raise TypeError(
                    f"cannot init from '{type(data).__name__}' of '{type(item).__name__}'"
                )

            result[k] = History(item.index).append(item)

        return result

    @staticmethod
    def _validateHistForFlags(history: History, colname=None):
        if history.empty:
            return history

        errm = f"History "
        if colname:
            errm += f"of column {colname} "

        # this ensures that the mask does not shadow UNFLAGGED with a NaN.
        if history.squeeze().hasnans:
            raise ValueError(errm + "is not valid (result of max() contains NaNs)")

        return history

    @property
    def _constructor(self) -> Type["Flags"]:
        return type(self)

    # ----------------------------------------------------------------------
    # meta data

    def keys(self) -> typing.KeysView:
        return self._data.keys()

    @property
    def columns(self) -> pd.Index:
        """
        Column index of the flags container

        Returns
        -------
        columns: pd.Index
            The columns index
        """
        return pd.Index(self._data.keys())

    @columns.setter
    def columns(self, value: pd.Index):
        """
        Set new columns names.

        Parameters
        ----------
        value : pd.Index
            New column names
        """
        if not isinstance(value, pd.Index):
            value = pd.Index(value)

        if not value.is_unique or not pd.api.types.is_string_dtype(value):
            raise TypeError("value must be pd.Index, with unique indices of type str")

        if not len(value) == len(self):
            raise ValueError("index must match current index in length")

        _data = {}

        for old, new in zip(self.columns, value):
            _data[new] = self._data[old]

        self._data = _data

    @property
    def empty(self) -> bool:
        """
        True if flags has no columns.

        Returns
        -------
        bool
            ``True`` if the container has no columns, otherwise ``False``.
        """
        return len(self._data) == 0

    def __len__(self) -> int:
        return len(self._data)

    def __contains__(self, item):
        return item in self.columns

    # ----------------------------------------------------------------------
    # item access

    @overload
    def __getitem__(self, key: str) -> pd.Series:
        ...

    @overload
    def __getitem__(self, key: list | pd.Index) -> Flags:
        ...

    def __getitem__(self, key: str | list | pd.Index) -> pd.Series | Flags:
        if isinstance(key, str):
            return self._data[key].squeeze()

        if isinstance(key, slice):
            key = self.columns[key]

        if isinstance(key, (list, pd.Index)):
            # only copy necessary data
            data = self._data
            try:
                self._data = {}
                new = self.copy()
            finally:
                self._data = data
            new._data = {k: self._data[k].copy() for k in key}
            return new

        raise TypeError(
            "Key must be of type str, list or index of string or slice,"
            f"not {type(key)}."
        )

    def __setitem__(self, key: SelectT, value: ValueT):
        # force-KW is only internally available

        if isinstance(key, tuple):
            if len(key) != 2:
                raise KeyError(
                    "a single 'column' or a tuple of 'mask, column' must be passt"
                )
            mask, key = key

            tmp = pd.Series(np.nan, index=self._data[key].index, dtype=float)

            # make a mask from an index, because it seems
            # that passing an index is a very common workflow
            if isinstance(mask, pd.Index):
                mask = pd.Series(True, index=mask, dtype=bool)
                mask = mask.reindex(tmp.index, fill_value=False)

            # raises (correct) KeyError
            try:
                if pd.api.types.is_list_like(value) and len(value) != len(tmp):
                    raise ValueError
                tmp[mask] = value
            except Exception:
                raise ValueError(
                    f"bad mask. cannot use mask of length {len(mask)} on "
                    f"data of length {len(tmp)}"
                )
            else:
                value = tmp

        # technically it would be possible to select a field and set
        # the entire column to a scalar flag value (float), but it has
        # a high potential, that this is not intended by the user.
        # if desired use ``flags[:, field] = flag``
        if not isinstance(value, pd.Series):
            raise ValueError(
                "expected a value of type 'pd.Series', "
                "if a scalar should be set, please use 'flags[:, field] = flag'"
            )

        if key not in self._data:
            self._data[key] = History(value.index)

        self._data[key].append(value, meta=None)

    def __delitem__(self, key):
        self._data.pop(key)

    def drop(self, key: str):
        """
        Delete a flags column.

        Parameters
        ----------
        key : str
            column name

        Returns
        -------
        flags object with dropped column, not a copy
        """
        self.__delitem__(key)

    # ----------------------------------------------------------------------
    # accessor

    @property
    def history(self) -> _HistAccess:
        """
        Accessor for the flags history.

        Access via ``flags.history['var']``.
        To set a new history use ``flags.history['var'] = value``.
        The passed value must be an instance of History or must be convertible to a
        history.

        Returns
        -------
        history : History
            Accessor for the flags history

        See Also
        --------
        saqc.core.History : History storage class.
        """
        return _HistAccess(self)

    # ----------------------------------------------------------------------
    # copy

    def copy(self, deep=True):
        """
        Copy the flags container.

        Parameters
        ----------
        deep : bool, default True
            If False, a new reference to the Flags container is returned,
            otherwise the underlying data is also copied.

        Returns
        -------
        copy of flags
        """
        new = self._constructor()
        new._data = {c: h.copy(deep) for c, h in self._data.items()}
        return new

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo=None):
        """
        Parameters
        ----------
        memo, default None
            Standard signature. Unused
        """
        return self.copy(deep=True)

    # ----------------------------------------------------------------------
    # transformation and representation

    def toDios(self) -> DictOfSeries:
        """
        Transform the flags container to a ``DictOfSeries``.


        .. deprecated:: 2.4
           use `saqc.DictOfSeries(obj)` instead.

        Returns
        -------
        DictOfSeries
        """
        warnings.warn(
            "toDios is deprecated, use `saqc.DictOfSeries(obj)` instead.",
            category=DeprecationWarning,
        )
        return DictOfSeries(self).copy()

    def toFrame(self) -> pd.DataFrame:
        """
        Transform the flags container to a ``pd.DataFrame``.

        Returns
        -------
        pd.DataFrame
        """
        return pd.DataFrame(dict(self))

    def __repr__(self) -> str:
        return str(DictOfSeries(self)).replace("DictOfSeries", type(self).__name__)


def initFlagsLike(
    reference: Union[pd.Series, DictLike, Flags],
    name: str = None,
) -> Flags:
    """
    Create empty Flags, from a reference data structure.

    Parameters
    ----------
    reference : pd.DataFrame, pd.Series, DictOfSeries, dict of pd.Series
        The reference structure to initialize for.

    name : str, default None
        Only respected if `reference` is of type ``pd.Series``.
        The column name that is used for the Flags. If ``None``
        the name of the series itself is taken, if it is unset,
        a ValueError is raised.

    Notes
    -----
    Implementation detail:

    The resulting Flags has not necessarily the exact same (inner) dimensions as the reference.
    This may happen, if the passed structure already holds History objects. Those are
    reduced 1D-DataFrame (1-column-History). Nevertheless the returned flags are perfectly suitable
    to be used in SaQC as flags container along with the passed reference structure (data).

    Returns
    -------
    flags: Flags
        a flags object,
    """
    result = {}

    if isinstance(reference, Flags):
        reference = reference._data

    if isinstance(reference, pd.Series):
        if name is None:
            name = reference.name
        if name is None:
            raise ValueError(
                "either the passed pd.Series must be named or a name must be passed"
            )
        if not isinstance(name, str):
            raise TypeError(f"name must be str not '{type(name).__name__}'")
        reference = reference.to_frame(name=name)

    for k, item in reference.items():
        if not isinstance(k, str):
            raise TypeError(
                f"cannot use '{k}' as a column name, currently only string keys are allowed"
            )
        if k in result:
            raise ValueError("reference must not have duplicate column names")
        if not isinstance(item, (pd.Series, History)):
            raise TypeError("items in reference must be of type pd.Series")

        result[k] = History(item.index)

    return Flags(result)
